# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import sys
import shutil
import json
import six
from subprocess import call
from optparse import OptionParser
from collections import OrderedDict
import logging
from distutils.dir_util import copy_tree


logger = logging.getLogger(__name__)

# Order of the different sections of the config.json. Any non-listed items will be placed at the end of the config
config_order = [
    'manifest',
    'target_simulator',
    'run',
    'conditions',
    'inputs',
    'components',
    'output',
    'reports',
    'networks'
]

network_dir_synonyms = ['network', 'networks', 'circuit', 'circuits', 'network_dir',  'circuit_dir']


class EnvBuilder(object):
    def __init__(self, base_dir='.', network_dir=None, components_dir=None, output_dir=None, node_sets_file=None):
        self._base_dir = self._get_base_dir(base_dir)
        self._network_dir = self._get_network_dir(network_dir)
        self._components_dir = self._get_components_dir(components_dir)
        self._output_dir = self._get_output_dir(output_dir)
        self._node_sets_file = self._get_node_sets_fname(node_sets_file)

        self._circuit_config = {}
        self._simulation_config = {}

    @property
    def target_simulator(self):
        raise NotImplementedError()

    @property
    def bmtk_simulator(self):
        raise NotImplementedError()

    @property
    def base_dir(self):
        return self._base_dir

    @property
    def scripts_root(self):
        local_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(local_path, 'scripts')

    @property
    def examples_dir(self):
        raise NotImplementedError()

    def _get_base_dir(self, base_dir):
        """Create base-dir, or if it exists make sure"""
        if not os.path.exists(base_dir):
            logger.info('Creating directory "{}"'.format(os.path.abspath(base_dir)))
            os.mkdir(base_dir)
        elif os.path.isfile(base_dir):
            logging.error('simulation directory path {} points to an existing file, cannot create.'.format(base_dir))
            exit(1)
        else:
            logging.info('Using existing directory "{}"'.format(os.path.abspath(base_dir)))

        return os.path.abspath(base_dir)

    @property
    def network_dir(self):
        return self._network_dir

    def _get_network_dir(self, network_dir=None):
        """Attempts to get the appropiate path the the directory containing the sonata network files, either as specified
        by the user (-n <network_dir> argument) or else attempt to find the location of the network files.

        :param base_dir:
        :param network_dir:
        :return: The absolute path of the directory that does/should contain the network files
        """
        network_dir_abs = None

        if network_dir is None:
            # Check to see if there are any folders in base_dir that might contain SONATA network files
            sub_dirs = [os.path.join(self.base_dir, dn) for dn in os.listdir(self.base_dir) if
                        os.path.isdir(os.path.join(self.base_dir, dn))
                        and dn.lower() in network_dir_synonyms]
            if sub_dirs:
                # See if there are any subfolders that might contain network files
                network_dir_abs = os.path.abspath(sub_dirs[0])
                logging.info(
                    'No network folder specified, attempting to use existing folder: {}'.format(network_dir_abs))
            else:
                network_dir_abs = os.path.abspath(os.path.join(self.base_dir, 'network'))
                logging.info('No network folder specified, creating empty folder: {}'.format(network_dir_abs))
                os.makedirs(network_dir_abs)

        elif os.path.isabs(network_dir):
            # In the case the user specifies the network directory as an absolute path
            network_dir_abs = network_dir
            logging.info('Using network directory: {}'.format(network_dir_abs))

        elif os.path.isdir(network_dir):
            # In case during calling the script the user was pointing to an already existing directory
            network_dir_abs = os.path.abspath(network_dir)
            logging.info('Using network directory: {}'.format(network_dir_abs))

        else:
            # Default case is the to use/add the <network_dir> found under base_dir
            network_dir_abs = os.path.join(self.base_dir, network_dir)
            if not os.path.exists(network_dir_abs):
                logging.info('Creating new network directory: {}'.format(network_dir_abs))
                os.makedirs(network_dir_abs)
            else:
                logging.info('Using network directory: {}'.format(network_dir_abs))

        return network_dir_abs

    def _parse_network_dir(self, network_dir):
        logger.info('Parsing {} for SONATA network files'.format(network_dir))
        net_nodes = {}
        net_edges = {}
        net_gaps = {}
        for root, dirs, files in os.walk(network_dir):
            for f in files:
                if not os.path.isfile(os.path.join(network_dir, f)) or f.startswith('.'):
                    continue

                if '_nodes' in f:
                    net_name = f[:f.find('_nodes')]
                    nodes_dict = net_nodes.get(net_name, {})
                    nodes_dict['nodes_file'] = os.path.abspath(os.path.join(root, f))
                    net_nodes[net_name] = nodes_dict
                    logger.info('  Adding nodes file: {}'.format(nodes_dict['nodes_file']))

                elif '_node_types' in f:
                    net_name = f[:f.find('_node_types')]
                    nodes_dict = net_nodes.get(net_name, {})
                    nodes_dict['node_types_file'] = os.path.abspath(os.path.join(root, f))
                    net_nodes[net_name] = nodes_dict
                    logger.info('  Adding node types file: {}'.format(nodes_dict['node_types_file']))

                elif '_edges' in f:
                    net_name = f[:f.find('_edges')]
                    edges_dict = net_edges.get(net_name, {})
                    edges_dict['edges_file'] = os.path.abspath(os.path.join(root, f))
                    net_edges[net_name] = edges_dict
                    logger.info('  Adding edges file: {}'.format(edges_dict['edges_file']))

                elif '_edge_types' in f:
                    net_name = f[:f.find('_edge_types')]
                    edges_dict = net_edges.get(net_name, {})
                    edges_dict['edge_types_file'] = os.path.abspath(os.path.join(root, f))
                    net_edges[net_name] = edges_dict
                    logger.info('  Adding edge types file: {}'.format(edges_dict['edge_types_file']))
                
                elif '_gap_juncs' in f:
                    net_name = f[:f.find('_gap_juncs')]
                    gaps_dict = net_gaps.get(net_name, {})
                    gaps_dict['gap_juncs_file'] = os.path.abspath(os.path.join(root, f))
                    net_gaps[net_name] = gaps_dict
                    logger.info('  Adding gap junctions file: {}'.format(gaps_dict['gap_juncs_file']))
                    
                else:
                    logger.info(
                        '  Skipping file (could not categorize): {}'.format(os.path.abspath(os.path.join(root, f))))

        if not (net_nodes or net_edges):
            logger.info('  Could not find any sonata nodes or edges file(s).')

        network_config = {'nodes': [], 'edges': [], 'gap_juncs': []}
        for _, sect in net_nodes.items():
            network_config['nodes'].append(sect)

        for _, sect in net_edges.items():
            network_config['edges'].append(sect)

        for _, sect in net_gaps.items():
            network_config['gap_juncs'].append(sect)

        self._circuit_config['networks'] = network_config

    @property
    def components_dir(self):
        return self._components_dir

    def _get_components_dir(self, components_dir=None):
        if components_dir is None:
            return os.path.abspath(os.path.join(self.base_dir, 'components'))

        elif os.path.isabs(components_dir):
            return components_dir

        elif os.path.exists(components_dir):
            return os.path.abspath(components_dir)

        else:
            return os.path.abspath(os.path.join(self.base_dir, components_dir))

    def _create_components_dir(self, components_dir, with_examples=True):
        if not os.path.exists(components_dir):
            logger.info('Creating components directory: {}'.format(components_dir))
            os.makedirs(components_dir)

        components_config = {}

        comps_dirs = [sd for sd in os.listdir(self.examples_dir) if os.path.isdir(os.path.join(self.examples_dir, sd))]
        for sub_dir in comps_dirs:
            comp_name = sub_dir + '_dir'
            src_dir = os.path.join(self.examples_dir, sub_dir)
            trg_dir = os.path.join(components_dir, sub_dir)
            if not os.path.exists(trg_dir):
                logger.info('Creating new components directory: {}'.format(trg_dir))
                os.makedirs(trg_dir)
            else:
                logger.info('Using components directory: {}'.format(trg_dir))

            components_config[comp_name] = trg_dir

            if with_examples:
                logger.info('  Copying files from {}.'.format(src_dir))
                copy_tree(src_dir, trg_dir)

        # return components_config
        self._circuit_config['components'] = components_config

    @property
    def output_dir(self):
        return self._output_dir

    def _get_output_dir(self, output_dir):
        if output_dir is None:
            return os.path.abspath(os.path.join(self.base_dir, 'output'))

        elif os.path.isabs(output_dir):
            return output_dir

        elif os.path.exists(output_dir):
            return os.path.abspath(output_dir)

        else:
            return os.path.abspath(os.path.join(self.base_dir, output_dir))

    def _add_manifest(self, config_dict, network_dir=None, components_dir=None, output_dir=None):
        config_dict['manifest'] = {'$BASE_DIR': '${configdir}'}
        base_dir = os.path.abspath(self.base_dir)

        replace_str = lambda fd, bd, var_name: fd.replace(bd, var_name) if fd.startswith(bd) else fd

        if network_dir is not None:
            config_dict['manifest']['$NETWORK_DIR'] = replace_str(network_dir, base_dir, '$BASE_DIR')
            if len(config_dict['networks'].get('nodes', [])) > 0:
                config_dict['networks']['nodes'] = [{k: replace_str(v, network_dir, '$NETWORK_DIR')
                                                     for k, v in l.items()} for l in config_dict['networks']['nodes']]

            if len(config_dict['networks'].get('edges', [])) > 0:
                config_dict['networks']['edges'] = [{k: replace_str(v, network_dir, '$NETWORK_DIR')
                                                     for k, v in l.items()} for l in config_dict['networks']['edges']]

        if 'network' in config_dict and isinstance(config_dict['network'], six.string_types):
            config_dict['network'] = replace_str(config_dict['network'], base_dir, '$BASE_DIR')

        if components_dir is not None:
            config_dict['manifest']['$COMPONENTS_DIR'] = replace_str(components_dir, base_dir, '$BASE_DIR')
            for k, v in config_dict['components'].items():
                config_dict['components'][k] = replace_str(v, components_dir, '$COMPONENTS_DIR')

        if output_dir is not None:
            config_dict['manifest']['$OUTPUT_DIR'] = replace_str(output_dir, base_dir, '$BASE_DIR')
            for k, v in config_dict.get('output', {}).items():
                if isinstance(v, six.string_types):
                    config_dict['output'][k] = replace_str(v, output_dir, '$OUTPUT_DIR')



        for input_dict in config_dict.get('inputs', {}).values():
            if 'input_file' in input_dict:
                input_dict['input_file'] = replace_str(input_dict['input_file'], base_dir, '$BASE_DIR')

            if 'file_name' in input_dict:
                input_dict['file_name'] = replace_str(input_dict['file_name'], base_dir, '$BASE_DIR')

        if 'node_sets_file' in config_dict:
            config_dict['node_sets_file'] = replace_str(config_dict['node_sets_file'], base_dir, '$BASE_DIR')


    def _save_config(self, json_dict, config_file_name):
        logger.info('Creating config file: {}'.format(config_file_name))
        with open(os.path.join(self.base_dir, config_file_name), 'w') as outfile:
            ordered_dict = OrderedDict(sorted(json_dict.items(),
                                              key=lambda s: config_order.index(s[0]) if s[0] in config_order else 100))
            json.dump(ordered_dict, outfile, indent=2)

    @property
    def node_sets_file(self):
        return self._node_sets_file

    def _get_node_sets_fname(self, node_sets_file):
        if node_sets_file is None or not os.path.isabs(node_sets_file):
            abs_path = os.path.abspath(os.path.join(self.base_dir, node_sets_file or 'node_sets.json'))
        else:
            abs_path = node_sets_file

        return abs_path

    def _create_node_sets_file(self, recorded_nodes=None, default_ns='all'):
        if os.path.exists(self.node_sets_file):
            logger.info('Found existing node sets file: {}'.format(self.node_sets_file))
        else:
            logger.info('Creating new node sets file: {}'.format(self.node_sets_file))
            node_sets = {
                'biophysical_nodes': {'model_type': 'biophysical'},
                'point_nodes': {'model_type': 'point_process'}
            }
            if recorded_nodes is not None:
                node_sets['recorded_nodes'] = {'node_ids': recorded_nodes} if isinstance(recorded_nodes, list) \
                    else {'population': recorded_nodes}

            json.dump(node_sets, open(self.node_sets_file, 'w'), indent=2)

        if recorded_nodes is not None:
            default_ns = 'recorded_nodes'

        self._simulation_config['node_sets_file'] = self.node_sets_file
        return default_ns

    def _add_reports(self, cell_vars, node_set, section='soma'):
        if isinstance(cell_vars, six.string_types):
            cell_vars = [cell_vars]

        for v in cell_vars:
            logger.info('Adding membrane report for variable {}'.format(v))

        report_config = {
            '{}_report'.format(v): {
                'variable_name': v,
                'cells': node_set,
                'module': 'membrane_report',
                'sections': section
            } for v in cell_vars}

        if 'reports' not in self._simulation_config:
            self._simulation_config['reports'] = {}

        self._simulation_config['reports'].update(report_config)

    def _add_clamp_reports(self, clamp_reports):
        if isinstance(clamp_reports, six.string_types):
            clamp_reports = [clamp_reports]

        for c in clamp_reports:
            logger.info("Adding clamp report for {} clamp.".format(c))

        report_config = {
            '{}_clamp_report'.format(c): {
                'variable_name': c,
                'module': 'clamp_report'
            } for c in clamp_reports}

        if 'reports' not in self._simulation_config:
            self._simulation_config['reports'] = {}

        self._simulation_config['reports'].update(report_config)

    def _add_output_section(self):
        output_section = {
            'log_file': 'log.txt',
            'output_dir': self.output_dir,
            'spikes_file': 'spikes.h5'
        }

        self._simulation_config['output'] = output_section

    def _add_current_clamp(self, current_param):
        if current_param is None:
            return

        logger.info('Adding current clamp')


        iclamp_config = {
            "input_type": "current_clamp",
            "module": "IClamp",
            "node_set": "all",
            "gids": current_param['gids'],
            "amp": current_param['amp'],
            "delay": current_param['delay'],
            "duration": current_param['duration']
        }

        if 'inputs' not in self._simulation_config:
            self._simulation_config['inputs'] = {}

        self._simulation_config['inputs']['current_clamp'] = iclamp_config

    def _add_file_current_clamp(self, current_param):
        if current_param is None:
            return

        amp_file = os.path.abspath(current_param["input_file"])

        f_iclamp_config = {
            "input_type": "current_clamp",
            "module": "FileIClamp",
            "node_set": "all",
            "input_file": amp_file
        }

        if 'inputs' not in self._simulation_config:
            self._simulation_config['inputs'] = {}

        self._simulation_config['inputs']['file_current_clamp'] = f_iclamp_config

    def _add_se_voltage_clamp(self, clamp_param):
        seclamp_config = {
            "input_type": "voltage_clamp",
            "module": "SEClamp",
            "node_set": "all",
            "gids": clamp_param['gids'],
            "amps": clamp_param['amps'],
            "durations": clamp_param["durations"]
        }

        if "rs" in clamp_param.keys():
            seclamp_config["rs"] = clamp_param["rs"]

        if 'inputs' not in self._simulation_config:
            self._simulation_config['inputs'] = {}

        name = 'se_voltage_clamp'
        if 'name' in clamp_param.keys():
            name = clamp_param['name']

        self._simulation_config['inputs'][name] = seclamp_config

    def _add_spikes_inputs(self, spikes_inputs):
        inputs_dict = {}
        for s in spikes_inputs:
            pop_name = s[0] or 'all'
            input_name = '{}_spikes'.format(s[0] or 'input')
            spikes_file = os.path.abspath(s[1])

            spikes_ext = os.path.splitext(spikes_file)[1][1:]
            spikes_ext = 'sonata' if spikes_ext in ['h5', 'hdf5'] else spikes_ext

            inputs_dict[input_name] = {
                "input_type": "spikes",
                "module": spikes_ext,
                "input_file": spikes_file,
                "node_set": pop_name
            }
        if 'inputs' not in self._simulation_config:
            self._simulation_config['inputs'] = {}

        self._simulation_config['inputs'].update(inputs_dict)

    def _add_run_params(self, tstart=0.0, tstop=1000.0, dt=0.001, **kwargs):
        self._simulation_config['run'] = {
            'tstart': tstart,
            'tstop': tstop,
            'dt': dt
        }

    def _copy_run_script(self):
        run_script = 'run_{}.py'.format(self.bmtk_simulator)
        shutil.copy(os.path.join(self.examples_dir, run_script), os.path.join(self.base_dir, run_script))

    def build(self, include_examples=False, use_relative_paths=True, report_vars=[],
              report_nodes=None, clamp_reports=[], current_clamp=None, file_current_clamp=None,
              se_voltage_clamp=None,
              spikes_inputs=None, config_file='config.json', overwrite_config=False, **run_args):

        config_path = config_file if os.path.isabs(config_file) else os.path.join(self._base_dir, config_file)
        config_filename = os.path.splitext(os.path.basename(config_file))[0]
        base_config = {
            'network': os.path.join(self._base_dir, 'circuit_' + config_filename + '.json'),
            'simulation': os.path.join(self._base_dir, 'simulation_' + config_filename + '.json')
        }
        if os.path.exists(config_path) and not overwrite_config:
            logging.warning(f'Configuration file {config_path} already exists, skipping. Please delete existing file, ' 
                            f'use a different name, or use overwrite_config=True.')
        else:
            self._save_config(base_config, config_path)

        self._parse_network_dir(self.network_dir)
        self._create_components_dir(self.components_dir, with_examples=include_examples)
        if use_relative_paths:
            self._add_manifest(self._circuit_config, network_dir=self.network_dir, components_dir=self.components_dir)
        if os.path.exists(base_config['network']) and not overwrite_config:
            logging.warning(f'Configuration file {base_config["network"]} already exists, skipping. '
                            f'Please delete existing file, use a different name, or use overwrite_config=True.')
        else:
            self._save_config(self._circuit_config, os.path.join(self._base_dir, 'circuit_'+config_filename+'.json'))

        selected_ns = self._create_node_sets_file(report_nodes)
        self._add_reports(report_vars, selected_ns)
        self._add_clamp_reports(clamp_reports)
        self._add_output_section()
        self._simulation_config['target_simulator'] = self.target_simulator
        self._add_run_params(**run_args)

        if current_clamp is not None:
            try:
                current_clamp['gids']
            except:
                current_clamp['gids']='all'

        self._add_current_clamp(current_clamp)

        if file_current_clamp is not None:
            self._add_file_current_clamp(file_current_clamp)

        if se_voltage_clamp is not None:
            try:
                se_voltage_clamp['gids']
            except:
                se_voltage_clamp['gids'] = 'all'

            self._add_se_voltage_clamp(se_voltage_clamp)

        if spikes_inputs!=None:
            self._add_spikes_inputs(spikes_inputs)
        if use_relative_paths:
            self._add_manifest(self._simulation_config, output_dir=self.output_dir)
        if os.path.exists(base_config['simulation']) and not overwrite_config:
            logging.warning(f'Configuration file {base_config["simulation"]} already exists, skipping. '
                            f'Please delete existing file, use a different name, or use overwrite_config=True.')
        else:
            self._save_config(self._simulation_config, os.path.join(self._base_dir, 'simulation_'+config_filename+'.json'))

        self._copy_run_script()


class BioNetEnvBuilder(EnvBuilder):
    @property
    def examples_dir(self):
        return os.path.join(self.scripts_root, 'bionet')

    @property
    def target_simulator(self):
        return 'NEURON'

    @property
    def bmtk_simulator(self):
        return 'bionet'

    def _add_run_params(self, tstart=0.0, tstop=1000.0, dt=0.001, dL=20.0, spikes_threshold=-15.0, nsteps_block=5000,
                        v_init=-80.0, celsius=34.0, **kwargs):
        self._simulation_config['run'] = {
            'tstart': tstart,
            'tstop': tstop,
            'dt': dt,
            'dL': dL,
            'spike_threshold': spikes_threshold,
            'nsteps_block': nsteps_block
        }

        self._simulation_config['conditions'] = {
            'celsius': celsius,
            'v_init': v_init
        }

    def compile_mechanisms(self):
        mechanisms_dir = os.path.join(self.components_dir, 'mechanisms')
        logger.info('Attempting to compile NEURON mechanims under "{}"'.format(mechanisms_dir))
        cwd = os.getcwd()

        try:
            os.chdir(os.path.join(mechanisms_dir))
            call(['nrnivmodl', 'modfiles'])
            logger.info('  Success.')
        except Exception as e:
            logger.error('  Was unable to compile mechanism in {}'.format(mechanisms_dir))
        os.chdir(cwd)


class PointNetEnvBuilder(EnvBuilder):
    @property
    def examples_dir(self):
        return os.path.join(self.scripts_root, 'pointnet')

    @property
    def target_simulator(self):
        return 'NEST'

    @property
    def bmtk_simulator(self):
        return 'pointnet'

    def _add_output_section(self):
        super(PointNetEnvBuilder, self)._add_output_section()
        self._simulation_config['output']['quiet_simulator'] = True


class FilterNetEnvBuilder(EnvBuilder):
    @property
    def examples_dir(self):
        return os.path.join(self.scripts_root, 'filternet')

    @property
    def target_simulator(self):
        return 'LGNModel'

    @property
    def bmtk_simulator(self):
        return 'filternet'

    def _add_output_section(self):
        super(FilterNetEnvBuilder, self)._add_output_section()
        self._simulation_config['output']['rates_file_csv'] = 'rates.csv'
        self._simulation_config['output']['spikes_file_csv'] = 'spikes.csv'
        self._simulation_config['output']['spikes_file_h5'] = 'spikes.h5'


class PopNetEnvBuilder(EnvBuilder):
    @property
    def examples_dir(self):
        return os.path.join(self.scripts_root, 'popnet')

    @property
    def target_simulator(self):
        return 'DiPDE'

    @property
    def bmtk_simulator(self):
        return 'popnet'

    def _add_output_section(self):
        super(PopNetEnvBuilder, self)._add_output_section()
        self._simulation_config['output']['rates_file'] = "firing_rates.csv"


def build_env_bionet(base_dir='.', network_dir=None, components_dir=None, node_sets_file=None, include_examples=False,
                     overwrite_config=False, tstart=0.0, tstop=1000.0, dt=0.001, dL=20.0, spikes_threshold=-15.0,
                     nsteps_block=5000, v_init=-80.0, celsius=34.0,
                     report_vars=[], report_nodes=None,
                     clamp_reports=[],
                     current_clamp=None,
                     file_current_clamp=None,
                     se_voltage_clamp=None,
                     spikes_inputs=None,
                     compile_mechanisms=False,
                     use_relative_paths=True,
                     config_file=None):
    env_builder = BioNetEnvBuilder(base_dir=base_dir, network_dir=network_dir, components_dir=components_dir,
                                   node_sets_file=node_sets_file)

    env_builder.build(include_examples=include_examples, use_relative_paths=use_relative_paths,
                      report_vars=report_vars, report_nodes=report_nodes, clamp_reports=clamp_reports,
                      current_clamp=current_clamp,
                      file_current_clamp=file_current_clamp, se_voltage_clamp=se_voltage_clamp, spikes_inputs=spikes_inputs,
                      tstart=tstart, tstop=tstop, dt=dt, dL=dL, spikes_threshold=spikes_threshold,
                      nsteps_block=nsteps_block, v_init=v_init, celsius=celsius, config_file=config_file,
                      overwrite_config=overwrite_config)

    if compile_mechanisms:
        env_builder.compile_mechanisms()


def build_env_pointnet(base_dir='.', network_dir=None, components_dir=None, node_sets_file=None, include_examples=False,
                       tstart=0.0, tstop=1000.0, dt=0.001, dL=20.0, spikes_threshold=-15.0, nsteps_block=5000,
                       v_init=-80.0, celsius=34.0,
                       report_vars=[], report_nodes=None, current_clamp=None,
                       spikes_inputs=None,
                       use_relative_paths=True,
                       config_file=None):
    env_builder = PointNetEnvBuilder(base_dir=base_dir, network_dir=network_dir, components_dir=components_dir,
                                     node_sets_file=node_sets_file)

    env_builder.build(include_examples=include_examples, use_relative_paths=use_relative_paths,
                      report_vars=report_vars, report_nodes=report_nodes, current_clamp=current_clamp,
                      spikes_inputs=spikes_inputs,
                      tstart=tstart, tstop=tstop, dt=dt, dL=dL, spikes_threshold=spikes_threshold,
                      nsteps_block=nsteps_block, v_init=v_init, celsius=celsius)


def build_env_filternet(base_dir='.', network_dir=None, components_dir=None,
                        node_sets_file=None, include_examples=False, tstart=0.0, tstop=1000.0, config_file='config.json'):
    env_builder = FilterNetEnvBuilder(base_dir=base_dir, network_dir=network_dir, components_dir=components_dir,
                                     node_sets_file=node_sets_file)

    env_builder.build(include_examples=include_examples,
                      base_dir=base_dir, network_dir=network_dir, components_dir=components_dir, tstart=tstart,
                      tstop=tstop, config_file=config_file)


def build_env_popnet(base_dir='.', network_dir=None, components_dir=None, node_sets_file=None, reports=None,
                     include_examples=True, config_file=None, tstart=0.0, tstop=1000.0, dt=0.001, **args):
    # raise NotImplementedError()

    env_builder = PopNetEnvBuilder(base_dir=base_dir, network_dir=network_dir, components_dir=components_dir,
                                     node_sets_file=node_sets_file)

    env_builder.build(include_examples=include_examples,
                      base_dir=base_dir, network_dir=network_dir, components_dir=components_dir, tstart=tstart,
                      tstop=tstop, config_file=config_file)

    # simulator='popnet'
    # target_simulator='DiPDE'
    # components_dir='pop_components'
    #
    # # Copy run script
    # copy_run_script(base_dir=base_dir, simulator=simulator, run_script='run_{}.py'.format(simulator))
    #
    # # Build circuit_config and componenets directory
    # circuit_config = build_circuit_env(base_dir=base_dir, network_dir=network_dir, components_dir=components_dir,
    #                                    simulator=simulator, with_examples=with_examples)
    # circuit_config['components']['population_models_dir'] = '$COMPONENTS_DIR/population_models'
    # # population_models_dir = os.path.join(base_dir, components_dir, 'population_models')
    # if with_examples:
    #     models_dir =  os.path.join(base_dir, components_dir, 'population_models')
    #     if os.path.exists(models_dir):
    #         shutil.rmtree(models_dir)
    #     shutil.copytree(os.path.join(scripts_path, simulator, 'population_models'), models_dir)
    #
    # copy_config(base_dir, circuit_config, 'circuit_config.json')
    #
    # # Build simulation config
    # simulation_config = build_simulation_env(base_dir=base_dir, target_simulator=target_simulator, tstop=tstop, dt=dt,
    #                                          reports=reports)
    # # PopNet doesn't produce spike files so instead need to replace them with rates files
    # for output_key in simulation_config['output'].keys():
    #     if output_key.startswith('spikes'):
    #         del simulation_config['output'][output_key]
    # # simulation_config['output']['rates_file_csv'] = 'firing_rates.csv'
    # simulation_config['output']['rates_file'] = 'firing_rates.csv'
    #
    # copy_config(base_dir, simulation_config, 'simulation_config.json')


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(module)s [%(levelname)s] %(message)s')

    parser = OptionParser(usage="Usage: python %prog [options] [bionet|pointnet|popnet|filternet] sim_dir")
    parser.add_option('-n', '--network_dir', dest='network_dir', default=None,
                      help="Use an exsting directory with network files.")
    parser.add_option('--tstop', type='float', dest='tstop', default=1000.0)
    parser.add_option('--dt', type=float, dest='dt', help='simulation time step dt', default=0.001)

    # For membrane report
    def membrane_report_parser(option, opt, value, parser):
        parser.values.has_membrane_report = True
        if ',' in value:
            try:
                setattr(parser.values, option.dest, [int(v) for v in value.split(',')])
            except ValueError as ve:
                setattr(parser.values, option.dest, value.split(','))

        else:
            setattr(parser.values, option.dest, value)

    parser.add_option('--report-vars', dest='mem_rep_vars', type='string', action='callback',
                      callback=membrane_report_parser, default=[],
                      help='A list of membrane variables to record from; v, cai, etc.')
    parser.add_option('--report-nodes', dest='mem_rep_cells', type='string', action='callback',
                      callback=membrane_report_parser, default=None)
    parser.add_option('--iclamp', dest='current_clamp', type='string', action='callback',
                      callback=membrane_report_parser, default=None,
                      help='Adds a soma current clamp using three variables: <amp>,<delay>,<duration> (nA, ms, ms)')
    parser.add_option('--spikes-inputs', dest='spikes_input', type='string', action='callback',
                      callback=membrane_report_parser, default=None)
    parser.add_option('--include-examples', dest='include_examples', action='store_true', default=False,
                      help='Copies component files used by examples and tutorials.')
    parser.add_option('--compile-mechanisms', dest='compile_mechanisms', action='store_true', default=False,
                      help='Will try to compile the NEURON mechanisms (BioNet only).')
    parser.add_option('--config-file', dest='config_file', type='string', default='config.json',
                      help='Name of configuration json file.')

    options, args = parser.parse_args()

    # Check the passed in argments are correct. [sim] </path/to/dir/>
    if len(args) < 2:
        parser.error('Invalid number of arguments, Please specify a target simulation (bionet, pointnet, filternet,'
                     'popnet) and the path to save the simulation environment.')
    elif len(args) > 2:
        parser.error('Unrecognized arguments {}'.format(args[2:]))
    else:
        target_sim = args[0].lower()
        if target_sim not in ['bionet', 'popnet', 'pointnet', 'filternet']:
            parser.error('Must specify one target simulator. options: "bionet", pointnet", "popnet", "filternet"')
        base_dir = args[1]

    if options.current_clamp is not None:
        cc_args = options.current_clamp
        if len(cc_args) != 3:
            parser.error('Invalid arguments for current clamp, requires three floating point numbers '
                         '<ampage>,<delay>,<duration> (nA, ms, ms)')
        iclamp_args = {'amp': float(cc_args[0]), 'delay': float(cc_args[1]), 'duration': float(cc_args[2])}
    else:
        iclamp_args = None

    spikes_inputs = []
    if options.spikes_input is not None:
        spikes = [options.spikes_input] if isinstance(options.spikes_input, str) else list(options.spikes_input)
        for spike_str in spikes:
            vals = spike_str.split(':')
            if len(vals) == 1:
                spikes_inputs.append((None, vals[0]))
            elif len(vals) == 2:
                spikes_inputs.append((vals[0], vals[1]))
            else:
                parser.error('Cannot parse spike-input string <pop1>:<spikes-file1>,<pop2>:<spikes-file2>,...')

    if target_sim == 'bionet':
        build_env_bionet(base_dir=base_dir, network_dir=options.network_dir, tstop=options.tstop,
                         dt=options.dt, report_vars=options.mem_rep_vars, report_nodes=options.mem_rep_cells,
                         current_clamp=iclamp_args, include_examples=options.include_examples,
                         spikes_inputs=spikes_inputs,
                         compile_mechanisms=options.compile_mechanisms,
                         config_file=options.config_file
                         )

    if target_sim == 'pointnet':
        build_env_pointnet(base_dir=base_dir, network_dir=options.network_dir, tstop=options.tstop,
                         dt=options.dt, report_vars=options.mem_rep_vars, report_nodes=options.mem_rep_cells,
                         current_clamp=iclamp_args, include_examples=options.include_examples,
                         spikes_inputs=spikes_inputs)

    elif target_sim == 'popnet':
        build_env_popnet(base_dir=base_dir, network_dir=options.network_dir, tstop=options.tstop,
                           dt=options.dt, config_file=options.config_file)

    elif target_sim == 'filternet':
        build_env_filternet(base_dir=base_dir, network_dir=options.network_dir, tstop=options.tstop,
                            include_examples=options.include_examples,
                            config_file=options.config_file)


# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
from bmtk.simulator.bionet.modules.sim_module import SimulatorMod
from bmtk.utils.reports.spike_trains import SpikeTrains, sort_order, sort_order_lu

from neuron import h


pc = h.ParallelContext()
MPI_RANK = int(pc.id())
N_HOSTS = int(pc.nhost())


class SpikesMod(SimulatorMod):
    """Module use for saving spikes

    """

    def __init__(self, tmp_dir, spikes_file_csv=None, spikes_file=None, spikes_file_nwb=None, cache_to_disk=True,
                 spikes_sort_order=None, mode='a'):
        # TODO: Have option to turn off caching spikes to csv.
        def _file_path(file_name):
            if file_name is None:
                return None

            if os.path.isabs(file_name):
                return file_name

            else:
                rel_tmp = os.path.realpath(tmp_dir)
                rel_fname = os.path.realpath(file_name)
                if not rel_fname.startswith(rel_tmp):
                    return os.path.join(tmp_dir, file_name)
                else:
                    return file_name

            # return file_name if os.path.isabs(file_name) else os.path.join(tmp_dir, file_name)

        self._csv_fname = _file_path(spikes_file_csv)
        self._save_csv = spikes_file_csv is not None

        self._h5_fname = _file_path(spikes_file)
        self._save_h5 = spikes_file is not None
        self._mode = mode

        self._nwb_fname = _file_path(spikes_file_nwb)
        self._save_nwb = spikes_file_nwb is not None

        self._tmpdir = tmp_dir
        self._sort_order = sort_order_lu.get(spikes_sort_order, sort_order.none)

        cache_name = os.path.basename(self._h5_fname or self._csv_fname or self._nwb_fname)
        self._spike_writer = SpikeTrains(cache_dir=tmp_dir, cache_name=cache_name, cache_to_disk=cache_to_disk)

        self._gid_map = None

    def initialize(self, sim):
        # TODO: since it's possible that other modules may need to access spikes, set_spikes_recordings() should
        # probably be called in the simulator itself.
        sim.set_spikes_recording()
        self._gid_map = sim.net.gid_pool

    def block(self, sim, block_interval):
        # take spikes from Simulator spikes vector and save to the tmp file
        for gid, tVec in sim.spikes_table.items():
            pop_id = self._gid_map.get_pool_id(gid)
            for t in tVec:
                self._spike_writer.add_spike(node_id=pop_id.node_id, timestamp=t, population=pop_id.population)

        pc.barrier()  # wait until all ranks have been saved
        sim.set_spikes_recording()  # reset recording vector

    def finalize(self, sim):
        # TODO: Get ride of flush/barrier calls, spike_trains should take care of it
        self._spike_writer.flush()
        pc.barrier()

        if self._save_csv:
            self._spike_writer.to_csv(self._csv_fname, sort_order=self._sort_order)
            pc.barrier()

        if self._save_h5:
            self._spike_writer.to_sonata(self._h5_fname, sort_order=self._sort_order, mode=self._mode)
            pc.barrier()

        if self._save_nwb:
            self._spike_writer.to_nwb(self._nwb_fname, sort_order=self._sort_order)
            pc.barrier()

        self._spike_writer.close()



from six import string_types
from bmtk.utils.sonata.config import SonataConfig as ConfigDict


def load_config(config):
    if isinstance(config, string_types):
        return ConfigDict.from_json(config)
    elif isinstance(config, dict):
        return ConfigDict.from_dict(config)
    else:
        raise Exception('Could not convert {} (type "{}") to json.'.format(config, type(config)))


import h5py
import numpy as np

from .core import CompartmentReaderABC
from bmtk.utils.hdf5_helper import get_attribute_h5


class _CompartmentPopulationReaderVer01(CompartmentReaderABC):
    sonata_columns = ['element_ids', 'element_pos', 'index_pointer', 'node_ids', 'time']

    def __init__(self, pop_grp, pop_name):
        self._pop_grp = pop_grp
        self._data_grp = pop_grp['data']
        self._mapping = pop_grp['mapping']
        self._population = pop_name

        self._gid2data_table = {}
        if self._mapping is None:
            raise Exception('could not find /mapping group')

        gids_ds = self._mapping[self.node_ids_ds]  # ['node_ids']
        index_pointer_ds = self._mapping['index_pointer']
        for indx, gid in enumerate(gids_ds):
            self._gid2data_table[gid] = slice(index_pointer_ds[indx], index_pointer_ds[indx+1])

        time_ds = self._mapping['time']
        self._t_start = float(time_ds[0])
        self._t_stop = float(time_ds[1])
        self._dt = float(time_ds[2])
        self._n_steps = int((self._t_stop - self._t_start) / self._dt)

        self._custom_cols = {col: grp for col, grp in self._mapping.items() if
                             col not in self.sonata_columns and isinstance(grp, h5py.Dataset)}

    def _get_index(self, node_id):
        if node_id not in self._gid2data_table:
            raise KeyError('node_id {} not found in {}/mapping/node_ids in file {}'.format(
                node_id, self._pop_grp.name, self._pop_grp.file.filename
            ))

        return self._gid2data_table[node_id]

    @property
    def populations(self):
        return [self._population]

    @property
    def data_ds(self):
        return self._data_grp

    @property
    def node_ids_ds(self):
        return 'node_ids'

    def get_population(self, population, default=None):
        raise NotImplementedError()

    def units(self, population=None):
        return get_attribute_h5(self.data_ds, 'units', None)
        #return self.data_ds.attrs.get('units', None)

    def variable(self, population=None):
        return get_attribute_h5(self.data_ds, 'variable', None)
        #return self.data_ds.attrs.get('variable', None)

    def tstart(self, population=None):
        return self._t_start

    def tstop(self, population=None):
        return self._t_stop

    def dt(self, population=None):
        return self._dt

    def n_steps(self, population=None):
        return self._n_steps

    def time(self, population=None):
        return self._mapping['time'][()]

    def time_trace(self, population=None):
        return np.linspace(self.tstart(), self.tstop(), num=self._n_steps, endpoint=True)

    def node_ids(self, population=None):
        return self._mapping['node_ids'][()]

    def index_pointer(self, population=None):
        return self._mapping['index_pointer'][()]

    def element_pos(self, node_id=None, population=None):
        if node_id is None:
            return self._mapping['element_pos'][()]
        else:
            return self._mapping['element_pos'][self._get_index(node_id)]#[indx_beg:indx_end]

    def element_ids(self, node_id=None, population=None):
        if node_id is None:
            return self._mapping['element_ids'][()]
        else:
            #indx_beg, indx_end = self._get_index(node_id)
            #return self._mapping['element_ids'][self._get_index(node_id)]#[indx_beg:indx_end]
            return self._mapping['element_ids'][self._get_index(node_id)]

    def n_elements(self, node_id=None, population=None):
        return len(self.element_pos(node_id))

    def data(self, node_id=None, population=None, time_window=None, sections='all', **opts):
        # filtered_data = self._data_grp
        multi_compartments = True
        if node_id is not None:
            node_range = self._get_index(node_id)
            if sections == 'origin' or self.n_elements(node_id) == 1:
                # Return the first (and possibly only) compartment for said gid
                gid_slice = node_range
                multi_compartments = False
            elif sections == 'all':
                # Return all compartments
                gid_slice = node_range #slice(node_beg, node_end)
            else:
                # return all compartments with corresponding element id
                compartment_list = list(sections) if np.isscalar(sections) else sections
                gid_slice = [i for i in self._get_index(node_id) if self._mapping['element_ids'] in compartment_list]
        else:
            gid_slice = slice(0, self._data_grp.shape[1])

        if time_window is None:
            time_slice = slice(0, self._n_steps)
        else:
            if len(time_window) != 2:
                raise Exception('Invalid time_window, expecting tuple [being, end].')

            window_beg = max(int((time_window[0] - self.tstart()) / self.dt()), 0)
            window_end = min(int((time_window[1] - self.tstart()) / self.dt()), self._n_steps / self.dt())
            time_slice = slice(window_beg, window_end)

        filtered_data = np.array(self._data_grp[time_slice, gid_slice])
        return filtered_data if multi_compartments else filtered_data[:]

    def custom_columns(self, population=None):
        return {k: v[()] for k,v in self._custom_cols.items()}

    def get_column(self, column_name, population=None):
        return self._mapping[column_name][()]

    def get_element_data(self, node_id, population=None):
        pass

    def get_report_description(self, population=None):
        pass

    def __getitem__(self, population):
        return self


class _CompartmentPopulationReaderVer00(_CompartmentPopulationReaderVer01):
    sonata_columns = ['element_id', 'element_pos', 'index_pointer', 'gids', 'time']

    def node_ids(self, population=None):
        return self._mapping[self.node_ids_ds][()]

    @property
    def node_ids_ds(self):
        return 'gids'

    def element_ids(self, node_id=None, population=None):
        if node_id is None:
            return self._mapping['element_id'][()]
        else:
            #indx_beg, indx_end = self._get_index(node_id)
            #return self._mapping['element_id'][self._get_index(node_id)]#[indx_beg:indx_end]
            return self._mapping['element_id'][self._get_index(node_id)]  # [indx_beg:indx_end]


class CompartmentReaderVer01(CompartmentReaderABC):
    def __init__(self, filename, mode='r', **params):
        self._h5_handle = h5py.File(filename, mode)
        self._h5_root = self._h5_handle[params['h5_root']] if 'h5_root' in params else self._h5_handle['/']
        self._popgrps = {}

        self._mapping = None

        if 'report' in self._h5_handle.keys():
            report_grp = self._h5_root['report']
            for pop_name, pop_grp in report_grp.items():
                self._popgrps[pop_name] = _CompartmentPopulationReaderVer01(pop_grp=pop_grp, pop_name=pop_name)
        else:
            self._default_population = 'pop_na'
            self._popgrps[self._default_population] = _CompartmentPopulationReaderVer00(pop_grp=self._h5_root, pop_name=self._default_population)

        if 'default_population' in params:
            # If user has specified a default population
            self._default_population = params['default_population']
            if self._default_population not in self._popgrps.keys():
                raise Exception('Unknown population {} found in report.'.format(self._default_population))
        elif len(self._popgrps.keys()) == 1:
            # If there is only one population in the report default to that
            self._default_population = list(self._popgrps.keys())[0]
        else:
            self._default_population = None

    @property
    def default_population(self):
        if self._default_population is None:
            raise Exception('Please specify a node population.')
        return self._default_population

    @property
    def populations(self):
        return list(self._popgrps.keys())

    def get_population(self, population, default=None):
        if population not in self.populations:
            return default
        return self[population]

    def units(self, population=None):
        population = population or self.default_population
        return self[population].units()

    def variable(self, population=None):
        population = population or self.default_population
        return self[population].variable()

    def tstart(self, population=None):
        population = population or self.default_population
        return self[population].tstart()

    def tstop(self, population=None):
        population = population or self.default_population
        return self[population].tstop()

    def dt(self, population=None):
        population = population or self.default_population
        return self[population].dt()

    def time_trace(self, population=None):
        population = population or self.default_population
        return self[population].time_trace()

    def node_ids(self, population=None):
        population = population or self.default_population
        return self[population].node_ids()

    def element_pos(self, node_id=None, population=None):
        population = population or self.default_population
        return self[population].element_pos(node_id)

    def element_ids(self, node_id=None, population=None):
        population = population or self.default_population
        return self[population].element_ids(node_id)

    def n_elements(self, node_id=None, population=None):
        population = population or self.default_population
        return self[population].n_elements(node_id)

    def data(self, node_id=None, population=None, time_window=None, sections='all', **opt_attrs):
        population = population or self.default_population
        return self[population].data(node_id=node_id, time_window=time_window, sections=sections, **opt_attrs)

    def custom_columns(self, population=None):
        population = population or self.default_population
        return self[population].custom_columns(population)

    def get_column(self, column_name, population=None):
        population = population or self.default_population
        return self[population].get_column(column_name)

    def get_node_description(self, node_id, population=None):
        population = population or self.default_population
        return self[population].get_node_description(node_id)

    def get_report_description(self, population=None):
        population = population or self.default_population
        return self[population].get_report_description()

    def __getitem__(self, population):
        return self._popgrps[population]



import os
import ast
import json
import types

import numpy as np


class SonataBaseEdge(object):
    def __init__(self, sonata_edge, edge_adaptor):
        self._edge = sonata_edge
        self._prop_adaptor = edge_adaptor

    @property
    def source_node_id(self):
        return self._edge.source_node_id

    @property
    def source_population(self):
        return self._edge.source_population

    @property
    def target_node_id(self):
        return self._edge.target_node_id

    @property
    def target_population(self):
        return self._edge.target_population

    @property
    def dynamics_params(self):
        return self._prop_adaptor.dynamics_params(self._edge)

    @property
    def delay(self):
        return self._edge['delay']

    @property
    def weight_function(self):
        return self._prop_adaptor.weight_function(self._edge)

    @property
    def preselected_targets(self):
        return self._prop_adaptor.preselected_targets

    @property
    def target_sections(self):
        return self._edge['target_sections']

    @property
    def target_distance(self):
        return self._edge['distance_range']

    @property
    def is_gap_junction(self):
        try:
            return self._edge['is_gap_junction']
        except:
            return False

    @property
    def edge_type_id(self):
        return self._edge.edge_type_id

    @property
    def nsyns(self):
        return self._prop_adaptor.nsyns(self._edge)

    @property
    def model_template(self):
        return self._edge['model_template']

    def syn_weight(self, src_node, trg_node):
        return self._prop_adaptor.syn_weight(self, src_node=src_node, trg_node=trg_node)

    def __getitem__(self, item):
        return self._edge[item]


class EdgeAdaptor(object):
    def __init__(self, network):
        self._network = network
        self._func_caches = self._network.py_function_caches

    @property
    def batch_process(self):
        return False

    @batch_process.setter
    def batch_process(self, flag):
        pass

    def get_edge(self, sonata_node):
        return SonataBaseEdge(sonata_node, self)

    @staticmethod
    def preprocess_edge_types(network, edge_population):
        edge_types_table = edge_population.types_table
        edge_type_ids = np.unique(edge_population.type_ids)

        for et_id in edge_type_ids:
            edge_type = edge_types_table[et_id]
            if 'dynamics_params' in edge_types_table.columns and edge_type['dynamics_params'] != None:
                dynamics_params = edge_type['dynamics_params']
                params_dir = network.get_component('synaptic_models_dir')

                params_path = os.path.join(params_dir, dynamics_params)

                # see if we can load the dynamics_params as a dictionary. Otherwise just save the file path and let the
                # cell_model loader function handle the extension.
                try:
                    params_val = json.load(open(params_path, 'r'))
                    edge_type['dynamics_params'] = params_val
                except Exception:
                    # TODO: Check dynamics_params before
                    network.io.log_exception('Could not find edge dynamics_params file {}.'.format(params_path))

            # Split target_sections
            if 'target_sections' in edge_type:
                trg_sec = edge_type['target_sections']
                if trg_sec is not None:
                    try:
                        edge_type['target_sections'] = ast.literal_eval(trg_sec)
                    except Exception as exc:
                        network.io.log_warning('Unable to split target_sections list {}'.format(trg_sec))
                        edge_type['target_sections'] = None

            # Split target distances
            if 'distance_range' in edge_type:
                dist_range = edge_type['distance_range']
                if dist_range is not None:
                    try:
                        # TODO: Make the distance range has at most two values
                        edge_type['distance_range'] = json.loads(dist_range)
                    except Exception as e:
                        try:
                            edge_type['distance_range'] = [0.0, float(dist_range)]
                        except Exception as e:
                            network.io.log_warning('Unable to parse distance_range {}'.format(dist_range))
                            edge_type['distance_range'] = None

    @classmethod
    def create_adaptor(cls, edge_group, network):
        prop_map = cls(network)
        return cls.patch_adaptor(prop_map, edge_group)

    @staticmethod
    def patch_adaptor(adaptor, edge_group):
        # dynamics_params
        if edge_group.has_dynamics_params:
            adaptor.dynamics_params = types.MethodType(group_dynamics_params, adaptor)
        else:  # 'dynamics_params' in node_group.all_columns:
            adaptor.dynamics_params = types.MethodType(types_dynamics_params, adaptor)

        # For fetching/calculating synaptic weights
        if 'weight_function' in edge_group.all_columns:
            # Customized function for user to calculate the synaptic weight
            adaptor.weight_function = types.MethodType(weight_function, adaptor)
            adaptor.syn_weight = types.MethodType(syn_weight_function, adaptor)
        elif 'syn_weight' in edge_group.all_columns:
            # Just return the synaptic weight
            adaptor.weight_function = types.MethodType(ret_none_function, adaptor)
            adaptor.syn_weight = types.MethodType(syn_weight, adaptor)
        else:
            raise Exception('Could not find syn_weight or weight_function properties. Cannot create connections.')

        # For determining the synapse placement
        if 'sec_id' in edge_group.all_columns or 'afferent_section_id' in edge_group.all_columns:
            adaptor.preselected_targets = True
            adaptor.nsyns = types.MethodType(no_nsyns, adaptor)
        elif 'nsyns' in edge_group.all_columns:
            adaptor.preselected_targets = False
            adaptor.nsyns = types.MethodType(nsyns, adaptor)
        else:
            # It will get here for connections onto point neurons
            adaptor.preselected_targets = True
            adaptor.nsyns = types.MethodType(no_nsyns, adaptor)

        return adaptor


def ret_none_function(self, edge):
    return None


def weight_function(self, edge):
    return edge['weight_function']


def syn_weight(self, edge, src_node, trg_node):
    return edge['syn_weight']


def syn_weight_function(self, edge, src_node, trg_node):
    weight_fnc_name = edge.weight_function
    if weight_fnc_name is None:
        weight_fnc = self._func_caches.py_modules.synaptic_weight('default_weight_fnc')
        return weight_fnc(edge, src_node, trg_node)

    elif self._func_caches.py_modules.has_synaptic_weight(weight_fnc_name):
        weight_fnc = self._func_caches.py_modules.synaptic_weight(weight_fnc_name)
        return weight_fnc(edge, src_node, trg_node)

    else:
        self._network.io.log_exception('weight_function {} is not defined.'.format(weight_fnc_name))


def nsyns(self, edge):
    return edge['nsyns']


def no_nsyns(self, edge):
    return 1


def types_dynamics_params(self, node):
    return node['dynamics_params']


def group_dynamics_params(self, node):
    return node.dynamics_params


# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import sys
import os
import glob
import neuron
from neuron import h

from bmtk.simulator.bionet.pyfunction_cache import py_modules
from bmtk.simulator.bionet.pyfunction_cache import load_py_modules
from bmtk.simulator.bionet.pyfunction_cache import synapse_model, synaptic_weight, cell_model
from bmtk.simulator.bionet.io_tools import io


pc = h.ParallelContext()


def quit_execution(): # quit the execution with a message
    pc.done()
    sys.exit()
    return


def clear_gids():
    pc.gid_clear()
    pc.barrier()


def load_neuron_modules(mechanisms_dir, templates_dir, default_templates=True):
    """

    :param mechanisms_dir:
    :param templates_dir:
    :param default_templates:
    """
    h.load_file('stdgui.hoc')

    bionet_dir = os.path.dirname(__file__)
    h.load_file(os.path.join(bionet_dir, 'import3d.hoc').replace("\\","/"))  # customized import3d.hoc to supress warnings
    h.load_file(os.path.join(bionet_dir,'default_templates',  'advance.hoc').replace("\\","/"))

    if isinstance(mechanisms_dir, list):
        for mdir in mechanisms_dir:
            try:
                neuron.load_mechanisms(str(mdir))
            except RuntimeError as rte:
                io.log_warning('Unable to load NEURON mechanisms.', display_once=True)

    elif mechanisms_dir is not None:
        try:
            neuron.load_mechanisms(str(mechanisms_dir))
        except RuntimeError as rte:
            io.log_warning('Unable to load NEURON mechanisms.', display_once=True)

    if default_templates:
        load_templates(os.path.join(bionet_dir, 'default_templates'))

    if templates_dir:
        load_templates(templates_dir)


def load_templates(template_dir):
    """Load all templates to be available in the hoc namespace for instantiating cells"""
    cwd = os.getcwd()
    os.chdir(template_dir)

    hoc_templates = glob.glob("*.hoc")

    for hoc_template in hoc_templates:
        h.load_file(str(hoc_template))

    os.chdir(cwd)


def reset():
    pc.gid_clear()
    pc.barrier()



# class CompartmentReport(object):
#     def __init__(self, path, mode='r', **kwargs):
#         pass
#
#     def initialize(self):
#         pass
#
#     def add_cell(self, node_id, sections, segments, population=None, **attrs):
#         pass
#
#     def record_cell(self, node_id, segment_vals, tstep, population=None):
#         pass
#
#     def record_cell_block(self, node_ids, segment_vals, tbegin, tend, population=None):
#         pass
#
#     def close(self):
#         pass
#
#     def flush(self):
#         pass
#
#     def from_sonata(self, path):
#         pass
#
#     def from_nwb(self, path):
#         pass
#
#
#     def from_path(self, path):
#         pass
#
#     def __getitem__(self, item):
#         pass


class CompartmentReaderABC(object):
    @property
    def populations(self):
        raise NotImplementedError()

    def get_population(self, population):
        raise NotImplementedError()

    def units(self, population=None):
        raise NotImplementedError()

    def variable(self, population=None):
        raise NotImplementedError()

    def tstart(self, population=None):
        raise NotImplementedError()

    def tstop(self, population=None):
        raise NotImplementedError()

    def dt(self, population=None):
        raise NotImplementedError()

    def time_trace(self, population=None):
        raise NotImplementedError()

    def n_steps(self, population=None):
        raise NotImplementedError()

    def node_ids(self, population=None):
        raise NotImplementedError()

    def element_pos(self, node_id=None, population=None):
        raise NotImplementedError()

    def element_ids(self, node_id=None, population=None):
        raise NotImplementedError()

    def n_elements(self, node_id, population=None):
        raise NotImplementedError()

    def index(self, population=None):
        raise NotImplementedError()

    def data(self, node_id=None, population=None, time_window=None, sections='all', **opt_attrs):
        raise NotImplementedError()

    def custom_columns(self, population=None):
        raise NotImplementedError()

    def get_column(self, column_name, population=None):
        raise NotImplementedError()

    def get_node_description(self, node_id, population=None):
        raise NotImplementedError()

    def get_report_description(self, population=None):
        raise NotImplementedError()

    def validate_file(self, path, **attrs):
        return False

    def __getitem__(self, population):
        return self.get_population(population)


class CompartmentWriterABC(object):
    def __init__(self, path, mode='r', **kwargs):
        raise NotImplementedError()

    def set_units(self, val, population=None):
        raise NotImplementedError()

    def set_variable(self, val, population=None):
        raise NotImplementedError()

    def set_tstart(self, val, population=None):
        raise NotImplementedError()

    def set_tstop(self, val, population=None):
        raise NotImplementedError()

    def set_dt(self, val, population=None):
        raise NotImplementedError()

    def set_time_trace(self, val, population=None):
        raise NotImplementedError()

    def initialize(self, **kwargs):
        raise NotImplementedError()

    def add_cell(self, node_id, element_ids, element_pos, population=None, **attrs):
        raise NotImplementedError()

    def record_cell(self, node_id, values, tstep, population=None):
        raise NotImplementedError()

    def record_cell_block(self, node_id, values, tbegin, tend, population=None):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def flush(self):
        raise NotImplementedError()

    def merge(self):
        raise NotImplementedError()

    def __getitem__(self, item):
        raise NotImplementedError()



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import h5py
import pandas as pd


class ColumnProperty(object):
    """Representation of a column name and metadata from a hdf5 dataset, csv column, etc.

    """
    def __init__(self, name, dtype, dimension, nrows=0, attrs=None):
        self._name = name
        self._dtype = dtype
        self._dim = dimension
        self._nrows = nrows
        self._attrs = attrs or {}

    @property
    def name(self):
        return self._name

    @property
    def dtype(self):
        return self._dtype

    @property
    def dimension(self):
        return self._dim

    @property
    def nrows(self):
        return self._nrows

    @property
    def attributes(self):
        return self._attrs

    @classmethod
    def from_h5(cls, hf_obj, name=None):
        if isinstance(hf_obj, h5py.Dataset):
            ds_name = name if name is not None else hf_obj.name.split('/')[-1]
            ds_dtype = hf_obj.dtype

            # If the dataset shape is in the form "(N, M)" then the dimension is M. If the shape is just "(N)" then the
            # dimension is just 1
            dim = 1 if len(hf_obj.shape) < 2 else hf_obj.shape[1]
            nrows = hf_obj.shape[0]
            return cls(ds_name, ds_dtype, dim, nrows, attrs=hf_obj.attrs)

        elif isinstance(hf_obj, h5py.Group):
            columns = []
            for name, ds in hf_obj.items():
                if isinstance(ds, h5py.Dataset):
                    columns.append(ColumnProperty.from_h5(ds, name))
            return columns

        else:
            raise Exception('Unable to convert hdf5 object {} to a property or list of properties.'.format(hf_obj))

    @classmethod
    def from_csv(cls, pd_obj, name=None):
        if isinstance(pd_obj, pd.Series):
            c_name = name if name is not None else pd_obj.name
            c_dtype = pd_obj.dtype
            return cls(c_name, c_dtype, 1)

        elif isinstance(pd_obj, pd.DataFrame):
            return [cls(name, pd_obj[name].dtype, 1) for name in pd_obj.columns]

        else:
            raise Exception('Unable to convert pandas object {} to a property or list of properties.'.format(pd_obj))

    def __hash__(self):
        return hash(self._name)

    def __repr__(self):
        return '{}'.format(self.name, self.dtype)

    def __eq__(self, other):
        if isinstance(other, ColumnProperty):
            return self._name == other._name
        else:
            return self._name == other



import os
import sys
import csv

import h5py
import pandas as pd
import numpy as np
from bmtk.utils.sonata.utils import add_hdf5_magic, add_hdf5_version


class SpikeTrainWriter(object):
    class TmpFileMetadata(object):
        def __init__(self, file_name, sort_order=None):
            self.file_name = file_name
            self.sort_order = sort_order

    def __init__(self, tmp_dir, mpi_rank=0, mpi_size=1):
        # For NEST/NEURON based simulations it is prefereable not to use mpi4py, so let the parent simulator determine
        # MPI rank and size
        self._mpi_rank = mpi_rank
        self._mpi_size = mpi_size

        # used to temporary save spike files since for large simulations saving spikes into memory can crash the
        # system. Requires the user to create the directory
        self._tmp_dir = tmp_dir
        if self._tmp_dir is None or not os.path.exists(self._tmp_dir):
            raise Exception('Directory path {} does not exists'.format(self._tmp_dir))
        self._all_tmp_files = [self.TmpFileMetadata(self._get_tmp_filename(r)) for r in range(mpi_size)]
        # TODO: Determine best buffer size.
        self._tmp_file_handle = open(self._all_tmp_files[mpi_rank].file_name, 'w')

        self._tmp_spikes_handles = []  # used when sorting mulitple file
        self._spike_count = -1

        # Nest gid files uses tab seperators and a different order for tmp spike files.
        self.delimiter = ' '  # delimiter for temporary file
        self.time_col = 0
        self.gid_col = 1

    def _get_tmp_filename(self, rank):
        return os.path.join(self._tmp_dir, '_bmtk_tmp_spikes_{}.csv'.format(rank))

    def _count_spikes(self, recount=False):
        if self._mpi_rank == 0:
            if self._spike_count > -1 and not recount:
                return self._spike_count

            self._spike_count = 0
            for tmp_file in self._all_tmp_files:
                with open(tmp_file.file_name, 'r') as csvfile:
                    csv_reader = csv.reader(csvfile, delimiter=self.delimiter)
                    self._spike_count += sum(1 for _ in csv_reader)

    def _sort_tmp_file(self, filedata, sort_order):
        # For now load spikes into pandas, it's the fastest way but may be an issue with memory
        if sort_order is None or filedata.sort_order == sort_order:
            return

        file_name = filedata.file_name
        tmp_spikes_ds = pd.read_csv(file_name, sep=' ', names=['time', 'gid'])
        tmp_spikes_ds = tmp_spikes_ds.sort_values(by=[sort_order, 'time'])
        tmp_spikes_ds.to_csv(file_name, sep=' ', index=False, header=False)
        filedata.sort_order = sort_order

    def _next_spike(self, rank):
        try:
            val = next(self._tmp_spikes_handles[rank])
            return val[0], val[1], rank
        except StopIteration:
            return None

    def add_spike(self, time, gid):
        self._tmp_file_handle.write('{:.6f} {}\n'.format(time, gid))

    def add_spikes(self, times, gid):
        for t in times:
            self.add_spike(t, gid)

    def add_spikes_file(self, file_name, sort_order=None):
        self._all_tmp_files.append(self.TmpFileMetadata(file_name, sort_order))

    def _sort_files(self, sort_order, sort_column, file_write_fnc):
        self._tmp_spikes_handles = []
        for fdata in self._all_tmp_files:
            self._sort_tmp_file(fdata, sort_order)
            self._tmp_spikes_handles.append(csv.reader(open(fdata.file_name, 'r'), delimiter=self.delimiter))

        spikes = []
        for rank in range(len(self._tmp_spikes_handles)):  # range(self._mpi_size):
            spike = self._next_spike(rank)
            if spike is not None:
                spikes.append(spike)

        # Iterate through all the ranks and find the first spike. Write that spike/gid to the output, then
        # replace that data point with the next spike on the selected rank
        indx = 0
        while spikes:
            # find which rank has the first spike
            selected_index = 0
            selected_val = spikes[0][sort_column]
            for i, spike in enumerate(spikes[1:]):
                if spike[sort_column] < selected_val:
                    selected_index = i + 1
                    selected_val = spike[sort_column]

            # write the spike to the file
            row = spikes.pop(selected_index)
            file_write_fnc(float(row[self.time_col]), int(row[self.gid_col]), indx)
            indx += 1

            # get the next spike on that rank and replace in spikes table
            another_spike = self._next_spike(row[2])
            if another_spike is not None:
                spikes.append(another_spike)

    def _merge_files(self, file_write_fnc):
        indx = 0
        for fdata in self._all_tmp_files:
            if not os.path.exists(fdata.file_name):
                continue

            with open(fdata.file_name, 'r') as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=self.delimiter)
                for row in csv_reader:
                    file_write_fnc(float(row[self.time_col]), int(row[self.gid_col]), indx)
                    indx += 1

    def _to_file(self, file_name, sort_order, file_write_fnc):
        if sort_order is None:
            sort_column = 0
        elif sort_order == 'time':
            sort_column = self.time_col
        elif sort_order == 'gid':
            sort_column = self.gid_col
        else:
            raise Exception('Unknown sort order {}'.format(sort_order))

        # TODO: Need to make sure an MPI_Barrier is called beforehand
        self._tmp_file_handle.close()
        if self._mpi_rank == 0:
            if sort_order is not None:
                self._sort_files(sort_order, sort_column, file_write_fnc)
            else:
                self._merge_files(file_write_fnc)

    def to_csv(self, csv_file, sort_order=None, gid_map=None):
        # TODO: Need to call flush and then barrier
        if self._mpi_rank == 0:
            # For the single rank case don't just copy the tmp-csv to the new name. It will fail if user calls to_hdf5
            # or to_nwb after calling to_csv.
            self._count_spikes()
            csv_handle = open(csv_file, 'w')
            csv_writer = csv.writer(csv_handle, delimiter=' ')

            def file_write_fnc_identity(time, gid, indx):
                csv_writer.writerow([time, gid])

            def file_write_fnc_transform(time, gid, indx):
                # For the case when NEURON/NEST ids don't match with the user's gid table
                csv_writer.writerow([time, gid_map[gid]])

            file_write_fnc = file_write_fnc_identity if gid_map is None else file_write_fnc_transform
            self._to_file(csv_file, sort_order, file_write_fnc)
            csv_handle.close()

        # TODO: Let user pass in in barrier and use it here

    def to_nwb(self, nwb_file):
        raise NotImplementedError

    def to_hdf5(self, hdf5_file, sort_order=None, gid_map=None):
        if self._mpi_rank == 0:
            with h5py.File(hdf5_file, 'w') as h5:
                add_hdf5_magic(h5)
                add_hdf5_version(h5)

                self._count_spikes(recount=True)
                spikes_grp = h5.create_group('/spikes')
                spikes_grp.attrs['sorting'] = 'none' if sort_order is None else sort_order
                time_ds = spikes_grp.create_dataset('timestamps', shape=(self._spike_count,), maxshape=(None,), dtype=np.float64)
                gid_ds = spikes_grp.create_dataset('gids', shape=(self._spike_count,), maxshape=(None,), dtype=np.uint64)

                def resize_data():
                    # There have been (unreproducable) mpi conditons where _count_spikes() is not correct, even with
                    # proper barriers and file flushing. Add a quick fix in case when converting csv to hdf5.
                    self._count_spikes(recount=True)
                    time_ds.resize((self._spike_count, ))
                    gid_ds.resize((self._spike_count, ))

                def file_write_fnc_identity(time, gid, indx):
                    if indx >= self._spike_count:
                        resize_data()
                    time_ds[indx] = time
                    gid_ds[indx] = gid

                def file_write_fnc_transform(time, gid, indx):
                    if indx >= self._spike_count:
                        resize_data()
                    time_ds[indx] = time
                    gid_ds[indx] = gid_map[gid]

                file_write_fnc = file_write_fnc_identity if gid_map is None else file_write_fnc_transform
                self._to_file(hdf5_file, sort_order, file_write_fnc)

        # TODO: Need to make sure a barrier is used here (before close is called)

    def flush(self):
        self._tmp_file_handle.flush()

    def close_tmp_file(self):
        self._tmp_file_handle.close()

    def close(self):
        if self._mpi_rank == 0:
            for tmp_file in self._all_tmp_files:
                if os.path.exists(tmp_file.file_name):
                    os.remove(tmp_file.file_name)


class PoissonSpikesGenerator(object):
    def __init__(self, gids, firing_rate, tstart=0.0, tstop=1000.0):
        self._gids = gids
        self._firing_rate = firing_rate / 1000.0
        self._tstart = tstart
        self._tstop = tstop

    def to_hdf5(self, file_name, sort_order='gid'):
        if sort_order == 'gid':
            gid_list = []
            times_list = []
            if sort_order == 'gid':
                for gid in self._gids:
                    c_time = self._tstart
                    while c_time < self._tstop:
                        interval = -np.log(1.0 - np.random.uniform()) / self._firing_rate
                        c_time += interval
                        gid_list.append(gid)
                        times_list.append(c_time)

            with h5py.File(file_name, 'w') as h5:
                h5.create_dataset('/spikes/gids', data=gid_list, dtype=np.uint)
                h5.create_dataset('/spikes/timestamps', data=times_list, dtype=np.float)
                h5['/spikes'].attrs['sorting'] = 'by_gid'

        else:
            raise NotImplementedError


class SpikesInput(object):
    def get_spikes(self, gid):
        raise NotImplementedError()

    @staticmethod
    def load(name, module, input_type, params):
        module_lc = module.lower()
        if module_lc == 'nwb':
            return SpikesInputNWBv1(name, module, input_type, params)
        elif module_lc == 'h5' or module_lc == 'hdf5':
            return SpikesInputH5(name, module, input_type, params)
        elif module_lc == 'csv':
            return SpikesInputCSV(name, module, input_type, params)
        else:
            raise Exception('Unable to load spikes for module type {}'.format(module))


class SpikesInputNWBv1(SpikesInput):
    def __init__(self, name, module, input_type, params):
        self.input_file = params['input_file']
        self._h5_handle = h5py.File(self.input_file, 'r')
        self.trial = params['trial']
        self._trial_grp = self._h5_handle['processing'][self.trial]['spike_train']

    def get_spikes(self, gid):
        return self._trial_grp[str(gid)]['data']


class SONATAIndexer(object):
    def get_spikes(self, gid):
        raise NotImplementedError()


class DictIndexedGIDs(SONATAIndexer):
    def __init__(self, spikes_input_h5):
        self._gid_indicies = {}
        self._parent = spikes_input_h5

        indx_beg = 0
        c_gid = self._parent.gids[0]
        for indx, gid in enumerate(self._parent.gids):
            # go through the gids dataset, determine slices for each gid
            if gid != c_gid:
                self._gid_indicies[c_gid] = slice(indx_beg, indx)
                c_gid = gid
                indx_beg = indx
        self._gid_indicies[c_gid] = slice(indx_beg, indx+1)  # saves the last entry

    def get_spikes(self, gid):
        if gid in self._gid_indicies:
            return self._parent.timestamps[self._gid_indicies[gid]]
        else:
            return []


class DFIndexedGIDs(SONATAIndexer):
    def __init__(self, spikes_input_h5):
        self._parent = spikes_input_h5

        index_df = pd.DataFrame(data={'gids': self._parent.gids}, )
        index_df.drop_duplicates(inplace=True)
        index_df = index_df.stack().reset_index()
        index_df.columns = ['indx_beg', 'tmp', 'gid']
        index_df.set_index('gid', inplace=True)
        index_df = index_df.drop('tmp', axis=1)
        index_df['indx_end'] = index_df['indx_beg'].shift(-1).fillna(len(self._parent.gids)).astype(np.int64)
        self._gid_indicies = index_df  # index_df.to_dict(orient='index')

    def get_spikes(self, gid):
        if gid in self._gid_indicies.index:
            indx_beg = self._gid_indicies.loc[gid]['indx_beg']
            indx_end = self._gid_indicies.loc[gid]['indx_end']
            return self._parent.timestamps[indx_beg:indx_end]
        else:
            return []


class HDF5IndexedGIDs(SONATAIndexer):
    def __init__(self, spikes_input_h5):
        raise NotImplementedError()


class UnindexedGIDs(SONATAIndexer):
    def __init__(self, spikes_input_h5):
        raise NotImplementedError()


class SpikesInputH5(SpikesInput):
    def __init__(self, name, module, input_type, params):
        self._input_file = params['input_file']
        self._h5_handle = h5py.File(self._input_file, 'r')
        self._gid_ds = self._h5_handle['/spikes/gids']
        self._timestamps_ds = self._h5_handle['/spikes/timestamps']
        self._sort_order = self._h5_handle['/spikes'].attrs.get('sorting', None)
        if sys.version_info[0] >= 3 and isinstance(self._sort_order, bytes) and self._sort_order is not None:
            # h5py attributes return str in py 2, bytes in py 3
            self._sort_order = self._sort_order.decode()

        # Create an index for fetching spike-trains based on gid
        if 'index' in self._h5_handle['/spikes'] and isinstance(self._h5_handle['/spikes/index'], h5py.Dataset):
            # In the case the index is built into the hdf5 file
            self._gid_index = HDF5IndexedGIDs(self)

        elif self._sort_order in ['gid', 'gids', 'by_gid']:
            # In the case when the sort_order == 'gid'
            # self._gid_index = DictIndexedGIDs(self)
            self._gid_index = DFIndexedGIDs(self)

        else:
            # In the case where the spike-trains are sorted by time or unsorted
            self._gid_index = UnindexedGIDs(self)

    @property
    def gids(self):
        return self._gid_ds

    @property
    def timestamps(self):
        return self._timestamps_ds

    def get_spikes(self, gid):
        return self._gid_index.get_spikes(gid)


class SpikesInputCSV(SpikesInput):
    def __init__(self, name, module, input_type, params):
        self._spikes_df = pd.read_csv(params['input_file'], index_col='gid', sep=' ')

    def get_spikes(self, gid):
        spike_times_str = self._spikes_df.iloc[gid]['spike-times']
        return np.array(spike_times_str.split(','), dtype=float)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import pandas as pd
import h5py
import numpy as np

from .utils import range_itr, get_attribute_h5
from .node import Node, NodeSet
from .edge import Edge, EdgeSet
from .group import NodeGroup, EdgeGroup


class Population(object):
    def __init__(self, pop_name, pop_group, types_table):
        self._pop_name = pop_name
        self._pop_group = pop_group
        self._types_table = types_table
        self._nrows = 0

        # For storing individual groups
        self._group_map = {}  # grp-id --> h5py.Group object
        self._find_groups()
        self._group_cache = {}  # grp-id --> soneta.io.Group() object

        # Refrences to most of the population's primary dataset
        self._type_id_ds = pop_group[self.type_ids_column]
        self._group_id_ds = pop_group[self.group_id_column]
        self._group_index_ds = pop_group[self.group_index_column]

        self._group_indicies = {}  # grp-id --> list of rows indicies
        self._group_indicies_cache_built = False

    @property
    def name(self):
        """name of current population"""
        return self._pop_name

    @property
    def group_ids(self):
        """List of all group_ids belonging to population"""
        return list(self._group_map.keys())

    @property
    def groups(self):
        """Returns a list of sonata.Group objects"""
        return [self.get_group(name) for name in self._group_map.keys()]

    @property
    def types_table(self):
        return self._types_table

    @property
    def type_ids(self):
        return np.array(self._type_id_ds)

    @property
    def group_id_ds(self):
        return self._group_id_ds

    @property
    def group_index_ds(self):
        return self._group_index_ds

    @property
    def group_id_column(self):
        raise NotImplementedError

    @property
    def group_index_column(self):
        raise NotImplementedError

    @property
    def type_ids_column(self):
        raise NotImplementedError

    def to_dataframe(self):
        """Convert Population to dataframe"""
        raise NotImplementedError

    def get_group(self, group_id):
        if group_id in self._group_cache:
            return self._group_cache[group_id]
        else:
            grp_h5 = self._group_map[group_id]
            grp_obj = self._build_group(group_id, grp_h5)
            self._group_cache[group_id] = grp_obj
            return grp_obj

    def group_indicies(self, group_id, build_cache=False):
        """Returns a list of all the population row index that maps onto the given group.

        Used for iterating or searching within a Group

        :param group_id: id of a given group
        :param build_cache: Will cache indicies for all groups. Will be faster if making multiple calls but requires
         more memory (default False)
        :return: A (possibly empty) list of row indicies (non-contiguous, but unique)
        """
        if self._group_indicies_cache_built:
            return self._group_indicies.get(group_id, [])

        else:
            tmp_index = pd.DataFrame()
            # TODO: Need to check the memory overhead, especially for edges. See if an iterative search is just as fast
            tmp_index['grp_id'] = pd.Series(self._group_id_ds[()], dtype=self._group_id_ds.dtype)
            tmp_index['row_indx'] = pd.Series(range_itr(self._nrows), dtype=np.uint32)
            if build_cache:
                # save all indicies as arrays
                self._group_indicies = {grp_id: np.array(subset['row_indx'])
                                        for grp_id, subset in tmp_index.groupby(by='grp_id')}
                self._group_indicies_cache_built = True
                return self._group_indicies.get(group_id, [])
            else:
                # TODO: Manually del tmp_index to clear out the memory?
                tmp_index = tmp_index[tmp_index['grp_id'] == group_id]
                return np.array(tmp_index['row_indx'])

    def igroup_ids(self, row_indicies):
        return self._group_id_ds[list(row_indicies)]

    def igroup_indicies(self, row_indicies):
        return self._group_index_ds[list(row_indicies)]

    def _find_groups(self):
        """Create a map between group-id and h5py.Group reference"""
        for grp_key, grp_h5 in self._pop_group.items():
            if grp_key.isdigit():
                grp_id = int(grp_key)
                self._group_map[grp_id] = grp_h5
            else:
                # TODO: Should we put a warning if an unrecognized group exists?
                pass

    def _build_group(self, group_id, group_h5):
        raise NotImplementedError

    def __len__(self):
        return self._nrows


class NodePopulation(Population):
    def __init__(self, pop_name, pop_group, node_types_tables):
        super(NodePopulation, self).__init__(pop_name=pop_name, pop_group=pop_group, types_table=node_types_tables)

        # TODO: node_ids can be implicit
        self._node_id_ds = pop_group['node_id']
        self._nrows = len(self._node_id_ds)

        # TODO: This isn't necessary if only using iterator. Delay building index until get_node() is called.
        self._index_nid2row = None  # A lookup from node_id --> h5 row number
        self._node_id_index_built = False
        self._build_node_id_index()

        # indicies for gid <--> node_id map
        self._has_gids = False
        self._index_gid2row = None  # gid --> row (for searching by gid)
        self._index_row2gid = None  # row --> gid (for iterator or searching by node-id)
        self._gid_lookup_fnc = lambda _: None  # for looking up gid by row, use fnc pointer rather than conditional

        self.__itr_index = 0  # for iterator

    @property
    def group_id_column(self):
        return 'node_group_id'

    @property
    def group_index_column(self):
        return 'node_group_index'

    @property
    def type_ids_column(self):
        return 'node_type_id'

    @property
    def has_gids(self):
        return self._has_gids

    @property
    def node_ids(self):
        return np.array(self._node_id_ds)

    @property
    def gids(self):
        if self.has_gids:
            return np.array(self._index_gid2row.index)
        else:
            return None

    @property
    def node_types_table(self):
        return self._types_table

    @property
    def index_column_name(self):
        return 'node_id'

    @property
    def node_types_table(self):
        return self.types_table

    def add_gids(self, gid_map_df, force=False):
        if self.has_gids and not force:
            # TODO: not sure if it's best to return an exception or just continue on in silence?
            raise Exception('Node population {} already has gids mapped onto node-ids.'.format(self.name))
            # return

        # Create map from gid --> node_id --> row #
        self._build_node_id_index()
        tmp_df = pd.DataFrame()
        tmp_df['row_id'] = self._index_nid2row.index
        tmp_df['node_id'] = self._index_nid2row
        gid_map_df = gid_map_df.merge(tmp_df, how='left', left_on='node_id', right_on='node_id')
        gid_map_df = gid_map_df.drop(['node_id', 'population'], axis=1)
        self._index_gid2row = gid_map_df.set_index('gid')
        self._index_row2gid = gid_map_df.set_index('row_id')
        self._gid_lookup_fnc = lambda row_indx: self._index_row2gid.loc[row_indx]['gid']
        self._has_gids = True

    def to_dataframe(self, index_by_id=True):
        if len(self.groups) == 1:
            ret_df = self.get_group(self.group_ids[0]).to_dataframe()
        else:
            ret_df = pd.DataFrame()
            for grp_id in self.group_ids:
                ret_df = ret_df.append(self.get_group(grp_id).to_dataframe(), sort=False)

        if index_by_id:
            ret_df = ret_df.set_index('node_id')

        return ret_df

    def get_row(self, row_indx):
        # TODO: Use helper function so we don't have to lookup gid/node_id twice
        # Note: I'm not cacheing the nodes for memory purposes, but it might be beneificial too.
        node_id = self._node_id_ds[row_indx]
        node_type_id = self._type_id_ds[row_indx]
        node_group_id = self._group_id_ds[row_indx]
        node_group_index = self._group_index_ds[row_indx]

        node_type_props = self.node_types_table[node_type_id]
        node_group_props = self.get_group(node_group_id)[node_group_index]
        node_gid = self._gid_lookup_fnc(row_indx)

        return Node(node_id, self._pop_name, node_type_id, node_type_props, node_group_id, node_group_props, None, gid=node_gid)

    def get_rows(self, row_indicies):
        """Returns a set of all nodes based on list of row indicies.

        Warning: currently due to the use of h5py, the list must be ordered and cannot contain duplicates.

        :param row_indicies: A list of row indicies
        :return: An iterable NodeSet of nodes in the specified indicies
        """
        # TODO: Check that row_indicies is unsigned and the max (which will be the last value) < n_rows
        # TODO: Check order and check for duplicates in list
        return NodeSet(row_indicies, self)

    def inode_ids(self, row_indicies):
        # You get errors if row_indicies is a numpy array or panda series so convert to python list
        # TODO: list conversion can be expensive, see if h5py will work with np arrays natively.
        return self._node_id_ds[list(row_indicies)]

    def igids(self, row_indicies):
        gids = self._gid_lookup_fnc(row_indicies)
        if gids is not None:
            gids = np.array(gids)
        return gids

    def inode_type_ids(self, row_indicies):
        # self._node_type_id_ds
        return self._type_id_ds[list(row_indicies)]

    def get_node_id(self, node_id):
        row_indx = self._index_nid2row.loc[node_id]
        return self.get_row(row_indx)

    def get_gid(self, gid):
        # assert(self.has_gids)
        row_indx = self._index_gid2row.loc[gid]['row_id']
        return self.get_row(row_indx)

    def filter(self, **filter_props):
        for grp in self.groups:
            for node in grp.filter(**filter_props):
                yield node

    def _build_node_id_index(self, force=False):
        if self._node_id_index_built and not force:
            return

        self._index_nid2row = pd.Series(range_itr(self._nrows), index=self._node_id_ds, dtype=self._node_id_ds.dtype)
        self._node_id_index_built = True

    def _build_group(self, group_id, group_h5):
        return NodeGroup(group_id, group_h5, self)

    def __iter__(self):
        self.__itr_index = 0
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self.__itr_index >= self._nrows:
            raise StopIteration

        nxt_node = self.get_row(self.__itr_index)
        self.__itr_index += 1
        return nxt_node

    def __getitem__(self, item):
        if isinstance(item, slice):
            # TODO: Check
            start = item.start if item.start is not None else 0
            stop = item.stop if item.stop is not None else self._nrows
            row_indicies = range_itr(start, stop, item.step)
            return NodeSet(row_indicies, self)

        elif isinstance(item, int):
            return self.get_row(item)

        elif isinstance(item, list):
            return NodeSet(item)
        else:
            print('Unable to get item using {}.'.format(type(item)))


class EdgePopulation(Population):
    class __IndexStruct(object):
        """Class sto store indicies subgroup"""
        # TODO: Use collections.namedtuple
        def __init__(self, lookup_table, edge_table):
            self.lookup_table = lookup_table
            self.edge_table = edge_table

    def __init__(self, pop_name, pop_group, edge_types_tables):
        super(EdgePopulation, self).__init__(pop_name=pop_name, pop_group=pop_group, types_table=edge_types_tables)

        # keep reference to source and target datasets
        self._source_node_id_ds = pop_group['source_node_id']
        self._target_node_id_ds = pop_group['target_node_id']

        self._nrows = len(self._source_node_id_ds)

        # TODO: Throw an error/warning if missing
        self._source_population = EdgePopulation.get_source_population(pop_group)
        self._target_population = EdgePopulation.get_target_population(pop_group)

        self.__itr_index = 0

        # TODO: use a function pointer for get_index so it doesn't have to run a conditional every time
        # TODO: add property and/or property so user can determine what indicies exists.
        self._targets_index = None
        self._has_target_index = False
        self._sources_index = None
        self._has_source_index = False
        self.build_indicies()

    @property
    def group_id_column(self):
        return 'edge_group_id'

    @property
    def group_index_column(self):
        return 'edge_group_index'

    @property
    def type_ids_column(self):
        return 'edge_type_id'

    @property
    def source_population(self):
        return self._source_population

    @property
    def target_population(self):
        return self._target_population

    @staticmethod
    def get_source_population(pop_group_h5):
        return get_attribute_h5(pop_group_h5['source_node_id'], 'node_population', None)

    @staticmethod
    def get_target_population(pop_group_h5):
        return get_attribute_h5(pop_group_h5['target_node_id'], 'node_population', None)

    @property
    def edge_types_table(self):
        return self._types_table

    def to_dataframe(self):
        raise NotImplementedError()


    def build_indicies(self):
        indicies_grp = None
        for grp_name in ['indices', 'indicies']:
            if grp_name in self._pop_group:
                indicies_grp = self._pop_group[grp_name]
                break

        # if 'indicies' in self._pop_group:
        if indicies_grp is not None:
            # indicies_grp = self._pop_group['indicies']
            for index_name, index_grp in indicies_grp.items():
                # TODO: Let __IndexStruct build the indicies
                # Make sure subgroup has the correct datasets
                if not isinstance(index_grp, h5py.Group):
                    continue

                if 'node_id_to_range' not in index_grp:
                    # TODO: make this more general, i.e 'id_to_range' thus we can index on gids, edge_types, etc
                    # TODO: Check that there are two columns in dataset
                    raise Exception('index {} in {} edges is missing column {}.'.format(index_name, self.name,
                                                                                        'node_id_to_range'))
                if 'range_to_edge_id' not in index_grp:
                    raise Exception('index {} in {} edges is missing column {}.'.format(index_name, self.name,
                                                                                        'range_to_edge_id'))

                # Cache the index
                targets_lookup = index_grp['node_id_to_range']
                edges_range = index_grp['range_to_edge_id']
                index_obj = self.__IndexStruct(targets_lookup, edges_range)

                # Determine the type of index
                if index_name == 'source_to_target':
                    self._sources_index = index_obj
                    self._has_source_index = True
                elif index_name == 'target_to_source':
                    self._targets_index = index_obj
                    self._has_target_index = True
                else:
                    # TODO: Need to send this to a logger rather than stdout
                    print('Unrecognized index {}. Ignoring.'.format(index_name))

    def _build_group(self, group_id, group_h5):
        return EdgeGroup(group_id, group_h5, self)

    def group_indicies(self, group_id, build_cache=False, as_list=False):
        if as_list:
            return super(EdgePopulation, self).group_indicies(group_id, build_cache)

        # For nodes it's safe to just keep a list of all indicies that map onto a given group. For edges bc there are
        # many more rows (and typically a lot less groups), We want to build an index like for source/target ids
        if len(self._group_map) == 1:
            return len(self), [[0, len(self)]]

        grp_indicies = super(EdgePopulation, self).group_indicies(group_id, build_cache=False)
        if len(grp_indicies) == 0:
            # Return an index with no ranges
            return 0, []

        # cluster into ranges. Naively implement, there is probably a faster way to cluster an ordered array!
        range_beg = grp_indicies[0]
        ranges = []
        for i in range_itr(1, len(grp_indicies)):
            if (grp_indicies[i-1]+1) != grp_indicies[i]:
                ranges.append([range_beg, grp_indicies[i-1]+1])
                range_beg = grp_indicies[i]
        ranges.append([range_beg, grp_indicies[-1]+1])
        return len(grp_indicies), np.array(ranges, dtype=np.uint32)

    '''
    def _get_target_index(self):
        # TODO: Do only once
        if self._targets_index is not None:
            return self._targets_index

        if 'incidies' in self._pop_group:
            if 'target_to_source' in self._pop_group['incidies']:
                targets_lookup = self._pop_group['incidies']['target_to_source']['node_id_to_range']
                edges_range = self._pop_group['incidies']['target_to_source']['range_to_edge_id']
                self._targets_index = self.__IndexStruct(targets_lookup, edges_range)
                return self._targets_index

        # TODO: What to do if index doesn't exist?
        raise NotImplementedError
    '''

    def get_row(self, index):
        src_node = self._source_node_id_ds[index]
        trg_node = self._target_node_id_ds[index]
        edge_type_id = self._type_id_ds[index]
        edge_types_props = self.edge_types_table[edge_type_id]

        edge_group_id = self._group_id_ds[index]
        edge_group_index = self._group_index_ds[index]
        edge_group_props = self.get_group(edge_group_id)[edge_group_index]
        return Edge(trg_node_id=trg_node, src_node_id=src_node, source_pop=self.source_population,
                    target_pop=self.target_population, group_id = edge_group_id,
                    group_props=edge_group_props, edge_types_props=edge_types_props)

    def filter(self, **filter_props):
        selected_edge_types = set(self.edge_types_table.edge_type_ids)
        types_filter = False  # Do we need to filter results by edge_type_id
        if 'edge_type_id' in filter_props:
            # TODO: Make sure the edge_type_id is valid
            selected_edge_types = set([filter_props['edge_type_id']])
            del filter_props['edge_type_id']
            types_filter = True

        if 'group_id' in filter_props:
            grp_id = filter_props['group_id']
            grp_id = [grp_id] if np.isscalar(grp_id) else grp_id
            selected_groups = set(grp_id)
            del filter_props['group_id']
        else:
            selected_groups = set(self._group_map.keys())  # list of grp_id's that will be used

        group_prop_filter = {}  # list of actual query statements
        group_filter = False  # do we need to filter results by group_id

        # Go through filter key==value pairs, create filters for groups and edge_types
        for filter_key, filter_val in filter_props.items():
            # Find out what groups, if any, the column should search in.
            group_query = False  # If it's querying a group property don't look in edge_types
            types_query = False
            for grp_id, grp_h5 in self._group_map.items():
                if filter_key in grp_h5:
                    # TODO: Need to check the dtype's match
                    selected_groups &= set([grp_id])
                    group_prop_filter[filter_key] = filter_val
                    group_query = True
                    group_filter = True

            if (not group_query) and filter_key in self.edge_types_table.columns:
                # Presearch the edge types and get only those edge_type_ids which match key==val
                selected_edge_types &= set(self.edge_types_table.find(filter_key, filter_val))
                types_filter = True
                types_query = True

            if not (group_query or types_query):
                # Property key neither exists in a group or the edge_types_table
                raise Exception('Could not find property {}'.format(filter_key))

        # Iterate through all nodes, only returning those that match the filter
        for indx in range_itr(self._nrows):
            # Filter by edge_type_id
            if types_filter:
                # TODO: Invert the selected_edge_types, it will be faster to fail immeditely than search the entire list
                if self._type_id_ds[indx] not in selected_edge_types:
                    continue

            # Filter by group properties
            if group_filter:
                # TODO: Invert group search
                grp_id = self._group_id_ds[indx]
                if grp_id not in selected_groups:
                    continue

                grp_index = self._group_index_ds[indx]
                search_failed = True
                for prop_key, prop_val in group_prop_filter.items():
                    if prop_val != self._group_map[grp_id][prop_key][grp_index]:
                        break
                else:
                    search_failed = False

                if search_failed:
                    continue

            yield self.get_row(indx)

    def get_target(self, target_node_id):
        # TODO: Raise an exception, or call find() and log a warning that the index is not available
        # TODO: check validity of target_node_id (non-negative integer and smaller than index range)
        assert(self._has_target_index)
        return self._get_index(self._targets_index, target_node_id)

    def get_targets(self, target_node_ids):
        # TODO: verify input is iterable
        assert(self._has_target_index)
        trg_index = self._targets_index
        for trg_id in target_node_ids:
            for edge in self._get_index(trg_index, trg_id):
                yield edge

    def get_source(self, source_node_id):
        assert(self._has_source_index)
        return self._get_index(self._sources_index, source_node_id)

    def get_sources(self, source_node_ids):
        assert(self._has_target_index)
        trg_index = self._sources_index
        for src_id in source_node_ids:
            for edge in self._get_index(trg_index, src_id):
                yield edge

    def _get_index(self, index_struct, lookup_id):
        # TODO: Use a EdgeSet instead
        if lookup_id >= len(index_struct.lookup_table):
            return []
            # raise StopIteration

        edges_table = index_struct.edge_table
        lookup_beg, lookup_end = index_struct.lookup_table[lookup_id]
        for i in range_itr(lookup_beg, lookup_end):
            edge_indx_beg, edge_indx_end = edges_table[i]
            for edge_indx in range_itr(edge_indx_beg, edge_indx_end):
                yield self.get_row(edge_indx)

    def __iter__(self):
        self.__itr_index = 0
        return self

    def __next__(self):
        if self.__itr_index >= self._nrows:
            raise StopIteration

        next_edge = self.get_row(self.__itr_index)
        self.__itr_index += 1
        return next_edge

    def next(self):
        return self.__next__()



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import pandas as pd
import numpy as np

try:
    from distutils.version import LooseVersion
    use_sort_values = LooseVersion(pd.__version__) >= LooseVersion('0.19.0')

except:
    use_sort_values = False


def spikes2dict(spikes_file):
    spikes_df = pd.read_csv(spikes_file, sep=' ', names=['time', 'gid'])

    if use_sort_values:
        spikes_sorted = spikes_df.sort_values(['gid', 'time'])
    else:
        spikes_sorted = spikes_df.sort(['gid', 'time'])

    spike_dict = {}
    for gid, spike_train in spikes_sorted.groupby('gid'):
        spike_dict[gid] = np.array(spike_train['time'])
    return spike_dict


def spike_files_equal(spikes_txt_1, spikes_txt_2, err=0.0001):
    trial_1 = spikes2dict(spikes_txt_1)
    trial_2 = spikes2dict(spikes_txt_2)
    if set(trial_1.keys()) != set(trial_2.keys()):
        return False

    for gid, spike_train1 in trial_1.items():
        spike_train2 = trial_2[gid]
        if len(spike_train1) != len(spike_train2):
            return False

        for s1, s2 in zip(spike_train1, spike_train2):
            if abs(s1 - s2) > err:
                return False

    return True


def get_mean_firing_rates(spike_gids, node_ids, tstop_msec):

    """
    Compute mean firing rate over the duration of the simulation
    
    :param spike_gids: gids of cells which spiked
    :param node_ids: np.array of node_ids

    :return mean_firing_rate: np.array mean firing rates

    """

    min_gid = np.min(node_ids)
    max_gid = np.max(node_ids)

    gid_bins = np.arange(min_gid-0.5,max_gid+1.5,1)
    hist,bins = np.histogram(spike_gids, bins=gid_bins)

    tstop_sec = tstop_msec*1E-3
    mean_firing_rates = hist/tstop_sec
    
    return mean_firing_rates



def spikes_equal_in_window(spikes1,spikes2,twindow):
    """
    Compare spikes within a time window    
    :param spikes1: dict with "time" and "gid" arrays for raster 1
    :param spikes2: dict with "time" and "gid" arrays for raster 2
    :param twindow: [tstart,tend] time window
    
    :return boolean: True if equal, False if different
    """

    ix1_window0=np.where(spikes1["time"]>twindow[0]) 
    ix1_window1=np.where(spikes1["time"]<twindow[1]) 
    ix1_window = np.intersect1d(ix1_window0,ix1_window1)


    ix2_window0=np.where(spikes2["time"]>twindow[0]) 
    ix2_window1=np.where(spikes2["time"]<twindow[1]) 
    ix2_window = np.intersect1d(ix2_window0,ix2_window1)

    print(len(spikes1["time"][ix1_window]),len(spikes2["time"][ix2_window]))
    if len(spikes1["time"][ix1_window]) != len(spikes2["time"][ix2_window]):
        print("There is a DIFFERENT number of spikes in each file within the window")
        print("No point to compare individual spikes")
        return
    else: 
        print("number of spikes are the same, checking details...")
    ix1_sort = np.argsort(spikes1["time"][ix1_window],kind="mergesort")
    ix2_sort = np.argsort(spikes2["time"][ix2_window],kind="mergesort")


    if (np.array_equal(spikes1["gid"][ix1_window[ix1_sort]],spikes2["gid"][ix2_window[ix2_sort]])) and (np.array_equal(spikes1["time"][ix1_window[ix1_sort]],spikes2["time"][ix2_window[ix2_sort]])):
        print("spikes are IDENTICAL!")
        return True
    else:
        print("spikes are DIFFERENT")
        return False




# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
from neuron import h
import h5py
import numpy as np

from bmtk.simulator.bionet.modules.sim_module import SimulatorMod
from bmtk.utils.reports.current_writer import CurrentWriterv01

pc = h.ParallelContext()
MPI_RANK = int(pc.id())
N_HOSTS = int(pc.nhost())


class ClampReport(SimulatorMod):
    def __init__(self, tmp_dir, file_name, variable_name, buffer_data=True, **kwargs):
        """Module used for saving NEURON clamp currents at each given step of the simulation.

        :param variable_name: which type of clamp it is a report of. As of now options are se, ic, and f_ic.
        :param clamps: indexes of the clamp list to make a report of.
        :param tmp_dir:
        :param file_name: name of h5 file to save variable to.
        :param buffer_data: Set to true then data will be saved to memory until written to disk during each block, reqs.
        more memory but faster. Set to false and data will be written to disk on each step (default: True)
        """

        self._tmp_dir = tmp_dir

        self._file_name = file_name if os.path.isabs(file_name) else os.path.join(tmp_dir, file_name)

        if N_HOSTS > 1:
            self._tmp_files = []
            for i in range(N_HOSTS):
                tmp_name = variable_name + str(i)
                self._tmp_files.append(os.path.join(self._tmp_dir, tmp_name))
            self._rank_file = self._tmp_files[MPI_RANK]
        else:
            self._rank_file = self._file_name

        self._var_recorder = None
        self._variable_name = variable_name

        self._buffer_data = buffer_data
    
    @property
    def variable(self):
        return self._variable_name

    def initialize(self, sim, clamps):
        self._clamps = clamps

        self._var_recorder = CurrentWriterv01(self._rank_file, num_currents=len(self._clamps),
                                              buffer_size=sim.nsteps_block, buffer_data=self._buffer_data, tstart=0.0,
                                              tstop=sim.tstop, dt=sim.dt)

        self._var_recorder.initialize()

    def step(self, sim, tstep):
        # save the current of each clamp at the current time-step.
        vals = []
        for clamp in self._clamps:
            vals.append(clamp._stim.i)
        self._var_recorder.record_clamps(vals, tstep)

    def block(self, sim, block_interval):
        # write variables in memory to file
        self._var_recorder.flush()

    def finalize(self, sim):
        # TODO: Build in mpi signaling into var_recorder
        pc.barrier()
        self._var_recorder.close()

        if MPI_RANK == 0:
            self.merge()

    def merge(self):
        if N_HOSTS > 1:
            h5final = h5py.File(self._file_name, 'w')
            all_currents = []
            for i in range(N_HOSTS):
                tmp_file = h5py.File(self._tmp_files[i], 'r')
                data = tmp_file['data'][()]
                for current in data:
                    if len(current) > 0:
                        all_currents.append(current)
                tmp_file.close()
                os.remove(self._tmp_files[i])
            h5final.create_dataset('data', data=np.array(all_currents))
            h5final.close()




from .network_adaptors.dm_network import DenseNetwork


"""
NetworkBuilder is just a slim container for the a NetworkAdaptor class object which does all the work of storing/saving
the network. Should make it easier to 
 1) Create/update the backend network-builder strategies on the fly by swapping out different network-adaptors
 2) Change the public interface/documentation without having to mess with the adaptor, see BioNetBuilder, 
    PointNetBuilder, etc.
"""


class NetworkBuilder(object):
    """The NetworkBuilder class is used for building and saving a brain network/circuit. By default it will save to
    SONATA format for running network simulations using BioNet, PointNet, PopNet or FilterNet bmtk modules, however it
    can be generalized to build any time of network for any time of simulation.

    Building the network:
        For the general use-case building a network consists of 4 steps, each with a corresponding method.

        1. Initialize the network::

            net = NetworkBuilder("network_name")

        2. Create nodes (ie cells) using the **add_nodes()** method::

            net.add_nodes(N=80, model_type='Biophysical', ei='exc')
            net.add_nodes(N=20, model_type='IntFire', ei='inh')
            ...

        3. Create Connection rules between different subsets of nodes using **add_edges()** method::

            net.add_edges(source={'ei': 'exc'}, target={'ei': 'inh'},
                          connection_rule=my_conn_func, synaptic_model='e2i')
            ...

        4. Finally **build** the network and **save** the files::

            net.build()
            net.save(output_dir='network_path')

        See the bmtk documentation, or the method doc-strings for more advanced functionality


    Network Accessor methods:
        **nodes()**

        Will return a iterable of Node objects for each node created. The Node objects can be used like dictionaries to
        fetch their properties. By default returns all nodes in the network, but you can filter out a given subset by
        passing in property/values pairs::

            for node in net.nodes(model_type='Biophysical', ei='exc'):
                assert(node['ei'] == 'exc')
                ...

        **edges()**

        Like the nodes() methods, but insteads returns a list of Edge type objects. It too can be filtered by an edge
        property::

            for edge in net.edges(synaptic_model='exp2'):
                ...

        One can also pass in a list of source (or target) to filter out only those edges which belong to a specific
        subset of cells::

            for edge in net.edges(target_nodes=net.nodes(ei='exc')):
            ...


    Network Properties:
        * **name** - name of the network
        * **nnodes** - number of nodes (cells) in the network.
        * **nedges** - number of edges. Will be zero if build() method hasn't been called
    """

    def __init__(self, name, adaptor_cls=DenseNetwork, **network_props):
        self.adaptor = adaptor_cls(name, **network_props)

    @property
    def name(self):
        """Get the name (string) of this network."""
        return self.adaptor.name

    @property
    def nodes_built(self):
        """Returns True if nodes has been instantiated for this network."""
        return self.adaptor.nodes_built

    @property
    def edges_built(self):
        """Returns True if the connectivity matrix has been instantiated for this network."""
        return self.adaptor.edges_built

    @property
    def nnodes(self):
        """Returns the number of nodes for this network."""
        return self.adaptor.nnodes

    @property
    def nedges(self):
        """Returns the total number of edges for this network."""
        return self.adaptor.nedges

    def add_nodes(self, N=1, **properties):
        """Used to add nodes (eg cells) to a network. User should specify the number of Nodes (N) and can use any
        properties/attributes they require to define the nodes. By default all individual cells will be assigned a
        unique 'node_id' to identify each node in the network and a 'node_type_id' to identify each group of nodes.

        If a property is a singluar value then said property will be shared by all the nodes in the group. If a value
        is a list of length N then each property will be uniquly assigned the each node. In the below example a group
        of 100 nodes is created, all share the same 'model_type' parameter but the pos_x values will be different for
        each node::

            net.add_nodes(N=100, pos_x=np.random.rand(100), model_type='intfire1', ...)

        You can use a tuple to store property values (in which the SONATA hdf5 will save it as a dataset with multiple
        columns). For example to have one property 'positions' which keeps track of the x/y/z coordinates of each cell::

            net.add_nodes(N=100, positions=[(rand(), rand(), rand()) for _ in range(100)], ...)

        :param N: number of nodes in this group
        :param properties: Individual and group properties of given nodes
        """
        self.adaptor.add_nodes(N=N, **properties)

    def nodes(self, **properties):
        """Returns an iterator of Node (glorified dictionary) objects, filtered by parameters.

        To get all nodes on a network::

            for node in net.nodes():
                ...

        To only get those nodes with properties that match a given list of parameter values::

            for nod in net.nodes(param1=value1, param2=value2, ...):
                ...

        :param properties: key-value pair of node attributes to filter returned nodes
        :return: An iterator of Node objects
        """
        return self.adaptor.nodes(**properties)

    def add_edges(self, source=None, target=None, connection_rule=1, connection_params=None, iterator='one_to_one',
                  **edge_type_properties):
        """Used to create the connectivity matrix between subsets of nodes. The actually connections will not be
        created until the build() method is called, using the 'connection_rule.

        Node Selection:
            To specify what subset of nodes will be used for the pre- and post-synaptic one can use a dictionary to
            filter the nodes. In the following all inh nodes will be used for the pre-synaptic neurons, but only exc
            fast-spiking neurons will be used in the post-synaptic neurons (If target or source is not specified all
            neurons will be used)::

                net.add_edges(source={'ei': 'inh'}, target={'ei': 'exc', 'etype': 'fast-spiking'},
                              dynamic_params='i2e.json', synaptic_model='alpha', ...)

            In the above code there is one connection between each source/target pair of nodes, but to create a
            multi-graph with N connections between each pair use 'connection_rule' parameter with an integer value::

                net.add_edges(
                    source={'ei': 'inh'},
                    target={'ei': 'exc', 'etype': 'fast-spiking'},
                    connection_rule=N,
                    ...
                )

        Connection rules:
            Usually the 'connection_rule' parameter will be the name of a function that takes in source-node and
            target-node object (which can be treated like dictionaries, and returns the number of connections (ie
            synapses, 0 or None if no synapses should exists) between the source and target cell::

                def my_conn_fnc(source_node, target_node):
                    src_pos = source_node['position']
                    trg_pos = target_node['position']
                    ...
                    return N_syns

                net.add_edges(source={'ei': 'exc'}, target={'ei': 'inh'}, connection_rule=my_conn_fnc, **opt_edge_attrs)

            If the connection_rule function requires addition arguments use the 'connection_params' option::

                def my_conn_fnc(source_node, target_node, min_edges, max_edges)
                    ...

                net.add_edges(connection_rule=my_conn_fnc, connection_params={'min_edges': 0, 'max_edges': 20}, ...)

            Sometimes it may be more efficient or even a requirement that multiple connections are created at the same
            time. For example a post-synaptic neuron may only be targeted by a limited number of sources which couldn't
            be done by the previous connection_rule function. But by setting property 'iterator' to value 'all_to_one'
            the connection_rule function now takes in as a value a list of N source neurons, a single target, and should
            return a list of size N::

                def bulk_conn_fnc(sources, target):
                    syn_list = np.zeros(len(sources))
                    for source in sources:
                        ....
                    return syn_list

                net.add_edges(connection_rule=bulk_conn_fnc, iterator='all_to_one', ...)

            There is also a 'one_to_all' iterator option that will pair each source node with a list of all available
            target nodes.

        Edge Properties:
            Normally the properties used when creating a given type of edge will be shared by all the individual
            connections. To create unique values for each edge, the add_edges() method returns a ConnectionMap object::

                def set_syn_weight_by_dist(source, target):
                    src_pos, trg_pos = source['position'], target['position']
                    ....
                    return syn_weight


                cm = net.add_edges(connection_rule=my_conn_fnc, model_template='Exp2Syn', ...)
                                delay=2.0)
                cm.add_properties('syn_weight', rule=set_syn_weight_by_dist)
                cm.add_properties('delay', rule=lambda *_: np.random.rand(0.01, 0.50))

            In this case the 'model_template' property has a value for all connections of this given type of edge. The
            'syn_weight' and 'delay' properties will (most likely) be unique values. See ConnectionMap documentation for
            more info.

        :param source: A dictionary or list of Node objects (see nodes() method). Used to filter out pre-synaptic
            subset of nodes.
        :param target: A dictionary or list of Node objects). Used to filter out post-synaptic subset of nodes
        :param connection_rule: Integer or a function that returns integer(s). Rule to determine number of connections
            between each source and target node
        :param connection_params: A dictionary, used when the 'connection_rule' is a function that requires additional
            argments
        :param iterator: 'one_to_one', 'all_to_one', 'one_to_all'. When 'connection_rule' is a function this sets
            how the subsets of source/target nodes are passed in. By default (one-to-one) the connection_rule is
            called for every source/target pair. 'all-to-one' will pass in a list of all possible source nodes for
            each target, and 'one-to-all' will pass in a list of all possible targets for each source.
        :param edge_type_properties: properties/attributes of the given edge type
        :return: A ConnectionMap object
        """
        return self.adaptor.add_edges(
            source=source,
            target=target,
            connection_rule=connection_rule,
            connection_params=connection_params,
            iterator=iterator,
            **edge_type_properties
        )

    def edges(self, target_nodes=None, source_nodes=None, target_network=None, source_network=None, **properties):
        """Returns a list of dictionary-like Edge objects, given filter parameters.

        To get all edges from a network::

            edges = net.edges()

        To specify the target and/or source node-set::

            edges = net.edges(target_nodes=net.nodes(type='biophysical'), source_nodes=net.nodes(ei='i'))

        To only get edges with a given edge_property::

          edges = net.edges(weight=100, syn_type='AMPA_Exc2Exc')

        :param target_nodes: gid, list of gid, dict or node-pool. Set of target nodes for a given edge.
        :param source_nodes: gid, list of gid, dict or node-pool. Set of source nodes for a given edge.
        :param target_network: name of network containing target nodes.
        :param source_network: name of network containing source nodes.
        :param properties: edge-properties used to filter out only certain edges.
        :return: list of bmtk.builder.edge.Edge properties.
        """
        return self.adaptor.edges(
            target_nodes=target_nodes,
            source_nodes=source_nodes,
            target_network=target_network,
            source_network=source_network,
            **properties
        )

    def build(self, force=False):
        """Builds nodes and edges.

        :param force: set true to force complete rebuilding of nodes and edges, if nodes() or save_nodes() has been
            called before then forcing a rebuild may change gids of each node.
        """
        self.adaptor.build(force=force)

    def save(self, output_dir='.', force_overwrite=True):
        """Used to save the network files in the appropriate (eg SONATA) format into the output_dir directory. The file
        names will be automatically generated based on the network names.

        To have more control over the output and file names use the **save_nodes()** and **save_edges()** methods.

        :param output_dir: string, directory where network files will be generated. Default, current working directory.
        :param force_overwrite: Overwrites existing network files.
        """
        self.adaptor.save(output_dir=output_dir, force_overwrite=force_overwrite)

    def save_nodes(self, nodes_file_name=None, node_types_file_name=None, output_dir='.', force_overwrite=True):
        """Save the instantiated nodes in SONATA format files.

        :param nodes_file_name: file-name of hdf5 nodes file. By default will use <network.name>_nodes.h5.
        :param node_types_file_name: file-name of the csv node-types file. By default will use
            <network.name>_node_types.csv
        :param output_dir: Directory where network files will be generated. Default, current working directory.
        :param force_overwrite: Overwrites existing network files.
        """
        self.adaptor.save_nodes(
            nodes_file_name=nodes_file_name,
            node_types_file_name=node_types_file_name,
            output_dir=output_dir,
            force_overwrite=force_overwrite
        )

    def save_edges(self, edges_file_name=None, edge_types_file_name=None, output_dir='.', src_network=None,
                   trg_network=None, name=None, force_build=True, force_overwrite=False):
        """Save the instantiated edges in SONATA format files.

        :param edges_file_name: file-name of hdf5 edges file. By default will use <src_network>_<trg_network>_edges.h5.
        :param edge_types_file_name: file-name of csv edge-types file. By default will use
            <src_network>_<trg_network>_edges.h5.
        :param output_dir: Directory where network files will be generated. Default, current working directory.
        :param src_network: Name of the source-node populations.
        :param trg_network: Name of the target-node populations.
        :param name: Name of edge populations, eg /edges/<name> in edges.h5 file.
        :param force_build: Force to (re)build the connection matrix if it hasn't already been built.
        :param force_overwrite: Overwrites existing network files.
        """
        self.adaptor.save_edges(
            edges_file_name=edges_file_name,
            edge_types_file_name=edge_types_file_name,
            output_dir=output_dir,
            src_network=src_network,
            trg_network=trg_network,
            name=name,
            force_build=force_build,
            force_overwrite=force_overwrite
        )

    def import_nodes(self, nodes_file_name, node_types_file_name):
        """Import nodes from an existing sonata file

        :param nodes_file_name:
        :param node_types_file_name:
        """
        self.adaptor.import_nodes(
            nodes_file_name=nodes_file_name,
            node_types_file_name=node_types_file_name
        )

    def clear(self):
        """Resets the network removing the nodes and edges created."""
        self.adaptor.clear()

    def __getattr__(self, item):
        """Catch-all for adaptor attributes"""
        return getattr(self.adaptor, item)


class BioNetBuilder(NetworkBuilder):
    """An instance of NetworkBuilder specifically designed for building BioNet (NEURON) complainant networks

    """
    def __init__(self, name, adaptor_cls=DenseNetwork, **network_props):
        super(BioNetBuilder, self).__init__(name=name, adaptor_cls=adaptor_cls, **network_props)

    def add_nodes(self, N=1, model_type=None, model_template=None, model_processing=None, dynamics_params=None,
                  morphology=None, x=None, y=None, z=None, rotation_angle_xaxis=None, rotation_angle_yaxis=None,
                  rotation_angle_zaxis=None, **properties):
        """Add a set of N NEURON type nodes/cells to the network

        :param N: number of nodes (cells) in this group
        :param model_type: the type of node, 'biophysical', 'virtual', 'single_compartment', or 'point_neuron'
        :param model_template: Template or class used to instantiate all instances of each cell, eg. a NeuroML or
            NEURON Hoc Template file. Should contain a prefix of the format type ('nml:<Cell.nml>' or
            'ctdb:Biophys1.hoc'
        :param model_processing: string, pre-defined or user function applied to instantiated cell property/morphology,
            if using Allen Cell-Type Database biophysical models use 'aibs_perisomatic' or 'aibs_allactive'.
        :param dynamics_params: A file-name or a dictionary of cell-dynamics parameter values used for generating the
            cell. If using models from Allen Cell-Type Database then passing in the name of <model>_fit.json file will
            apply the same model params to all N cells. If you want unique params for each cell pass in a dictionary
            where each key has an associated list of size N::

                dynamics_params={'g_bar': [0.5, 0.2, ...], 'na_bar': [1.653e-5, 9.235e-6, ...]}

        :param morphology: string, name of morphology swc file used for creating model_type=biophysical cells
        :param x: list/array of size N floats for each cell's x-axis soma position
        :param y: list/array of size N floats for each cell's y-axis soma position
        :param z: list/array of size N floats for each cell's z-axis soma position
        :param rotation_angle_xaxis: x-axis euler rotation angle around the soma in radians
        :param rotation_angle_yaxis: y-axis euler rotation angle around the soma in radians
        :param rotation_angle_zaxis: z-axis euler rotation angle around the soma in radians
        :param properties: Individual and group properties of given nodes
        """
        for arg_var, arg_val in locals().items():
            if arg_var in ['properties', 'self', 'N']:
                continue
            elif arg_val is not None:
                properties[arg_var] = arg_val

        super(BioNetBuilder, self).add_nodes(N=N, **properties)

    def add_edges(self, source=None, target=None, connection_rule=1, connection_params=None, iterator='one_to_one',
                  model_template=None, dynamics_params=None, syn_weight=None, delay=None, target_sections=None,
                  distance_range=None, **properties):
        """Add rules for creating edges between a subset of source and target cells.

        :param source: A dictionary or list of Node objects (see nodes() method). Used to filter out pre-synaptic
            subset of nodes.
        :param target: A dictionary or list of Node objects). Used to filter out post-synaptic subset of nodes
        :param connection_rule: Integer or a function that returns integer(s). Rule to determine number of connections
            between each source and target node
        :param connection_params: A dictionary, used when the 'connection_rule' is a function that requires additional
            argments
        :param iterator: 'one_to_one', 'all_to_one', 'one_to_all'. When 'connection_rule' is a function this sets
            how the subsets of source/target nodes are passed in. By default (one-to-one) the connection_rule is
            called for every source/target pair. 'all-to-one' will pass in a list of all possible source nodes for
            each target, and 'one-to-all' will pass in a list of all possible targets for each source.
        :param model_template: A predefined or user generated NEURON function name for generating connection. (default:
            exp2syn).
        :param dynamics_params: A json file path or a dictionary of edge/synapse dynamics parameter values used for
            generating the connection, according to the model_template function. If passing in the name or path of a
            json file the parameter values will be applied to a edges.
        :param syn_weight: float or list of N floats, weights applied to each edge/synpase, value dependent on model
        :param delay: float or list of N floats, delay in milliseconds for cell-cell events
        :param target_sections: list of strings, non-SONATA compliant, locations of where to place synpases when the
            target cell is model_type=biophysical. One or more of the following values: somatic, basal, apical, axon.
            If specified post-synaptic location will be randomly chosen from given sections (+ distance_range value).
            To assign specific locations use afferent_section_id/afferent_section_pos along with the ConnectionMap
            instance returned.
        :param distance_range: A numeric range of floats [beg, end]. non SONATA compliant. Start and end arc-length
            distance from SOMA where the post-synaptic location will be placed on a model_type=biophysical cell. Along
            with target_sections parameter can be used to randomly but strategically connection cells. To assign
            specific locations use afferent_section_id/afferent_section_pos along with the returned ConnectionMap.
        :param properties: properties/attributes of the given edge type
        :return: A ConnectionMap object
        """
        for arg_var, arg_val in locals().items():
            if arg_var in ['properties', 'source', 'target', 'connection_rule', 'connection_params', 'iterator']:
                continue
            elif arg_val is not None:
                properties[arg_var] = arg_val

        return super(BioNetBuilder, self).add_edges(
            source=source,
            target=target,
            connection_rule=connection_rule,
            connection_params=connection_params,
            iterator=iterator,
            **properties
        )


class PointNetBuilder(NetworkBuilder):
    """An instance of NetworkBuilder specifically designed for building PointNet (NEST) complainant networks

    """
    def __init__(self, name, adaptor_cls=DenseNetwork, **network_props):
        super(PointNetBuilder, self).__init__(name=name, adaptor_cls=adaptor_cls, **network_props)

    def add_nodes(self, N=1, model_type='point_neuron', model_template=None, dynamics_params=None, x=None, y=None,
                  z=None, **properties):
        """Add a set of N NEST model nodes/cells to the network.

        :param N:
        :param model_type: The type of node, should be value 'point_neuron'
        :param model_template: Name of NEST neuron model to used, should have a 'nest:' prefix; nest:iaf_psc_alpha,
            nest:glif_asc
        :param dynamics_params: A file-name or a dictionary of cell-dynamics parameter values used for generating the
            cell, with parameters matching that of the parameters used to initialize NEST model_template cell. If you
            want unique params for each cell pass in a dictionary where each key has an associated list of size N::
                
                dynamics_params={'v_init': [-90, -85, ...], 'v_reset': [-20, -20, ...]}
                
        :param x: list/array of size N floats for each cell's x-axis position
        :param y: list/array of size N floats for each cell's y-axis position
        :param z: list/array of size N floats for each cell's z-axis position
        :param properties: Individual and group properties of given nodes
        """
        for arg_var, arg_val in locals().items():
            if arg_var in ['properties', 'self', 'N']:
                continue
            elif arg_val is not None:
                properties[arg_var] = arg_val

        super(PointNetBuilder, self).add_nodes(N=N, **properties)

    def add_edges(self, source=None, target=None, connection_rule=1, connection_params=None, iterator='one_to_one',
                  model_template=None, dynamics_params=None, syn_weight=None, delay=None, **properties):
        """Add rules for creating edges between a subset of source and target cells.

        :param source: A dictionary or list of Node objects (see nodes() method). Used to filter out pre-synaptic
            subset of nodes.
        :param target: A dictionary or list of Node objects). Used to filter out post-synaptic subset of nodes
        :param connection_rule: Integer or a function that returns integer(s). Rule to determine number of connections
            between each source and target node
        :param connection_params: A dictionary, used when the 'connection_rule' is a function that requires additional
            argments
        :param iterator: 'one_to_one', 'all_to_one', 'one_to_all'. When 'connection_rule' is a function this sets
            how the subsets of source/target nodes are passed in. By default (one-to-one) the connection_rule is
            called for every source/target pair. 'all-to-one' will pass in a list of all possible source nodes for
            each target, and 'one-to-all' will pass in a list of all possible targets for each source.
        :param model_template: A predefined or user generated NEURON function name for generating connection. (default:
            exp2syn).
        :param dynamics_params: A json file path or a dictionary of edge/synapse dynamics parameter values used for
            generating the connection, according to the model_template function. If passing in the name or path of a
            json file the parameter values will be applied to a edges.
        :param syn_weight: float or list of N floats, weights applied to each edge/synpase, value dependent on model
        :param delay: float or list of N floats, delay in milliseconds for cell-cell events
        :param properties: properties/attributes of the given edge type
        :return: A ConnectionMap object
        """
        for arg_var, arg_val in locals().items():
            if arg_var in ['properties', 'source', 'target', 'connection_rule', 'connection_params', 'iterator']:
                continue
            elif arg_val is not None:
                properties[arg_var] = arg_val

        return super(PointNetBuilder, self).add_edges(
            source=source,
            target=target,
            connection_rule=connection_rule,
            connection_params=connection_params,
            iterator=iterator,
            **properties
        )


class PopNetBuilder(NetworkBuilder):
    """An instance of NetworkBuilder specifically designed for building PopNet (DiPDE) complainant networks

    """
    def __init__(self, name, adaptor_cls=DenseNetwork, **network_props):
        super(PopNetBuilder, self).__init__(name=name, adaptor_cls=adaptor_cls, **network_props)

    def add_nodes(self, model_type='population', model_template=None, dynamics_params=None, **properties):
        """Add a DiPDE cell population

        :param model_type: The type of node, should be value 'population'
        :param model_template: Should be either 'dipde:Internal' or 'dipde:External' (or 'virtual')
        :param dynamics_params: path to json file, or dictionary, of parameters for dipde population
        :param properties: Individual and group properties of given nodes
        """
        for arg_var, arg_val in locals().items():
            if arg_var in ['properties', 'self']:
                continue
            elif arg_val is not None:
                properties[arg_var] = arg_val

        super(PopNetBuilder, self).add_nodes(N=1, **properties)

    def add_edges(self, source=None, target=None, nsyns=1, syn_weight=None, delay=None, dynamics_params=None,
                  **properties):
        """

        :param source: A dictionary or list of efferent dipde Populations
        :param target: A dictionary or list of afferent dipde Populations
        :param nsyns: Number of connections between the populations
        :param syn_weight: connection weight between the populations
        :param delay: connection delay between the populations
        :param dynamics_params: Additional parameters used for initializing connection
        :param properties: properties/attributes of the given edge type
        :return: A ConnectionMap object
        """
        for arg_var, arg_val in locals().items():
            if arg_var in ['properties', 'source', 'target']:
                continue
            elif arg_val is not None:
                properties[arg_var] = arg_val

        return super(PopNetBuilder, self).add_edges(
            source=source,
            target=target,
            **properties
        )
